exception InputHelpError of string

(** path of the problem file *)
let problemfile = ref ""

(** arguments to be passed to sub-processes (i.e. "slaves") *)
let slave = ref false
let slaveargs = ref [Sys.argv.(0)];;
let preparsed = ref false

(** setting multiple flags simultaneously *)
let mode : string list ref = ref []
let schedule_files : string list ref = ref []
let flag_overrides : (string * string) list ref = ref []

(** criteria for stopping proof search *)
let     timeout : float option ref = ref None
let hardtimeout : float option ref = ref None
let max_searchoptions : int option ref = ref None

(** Set completep to false if I use a mode that makes the search incomplete,
    so that failure does not imply satisfiability. *)
let completep = ref true
(** Know if you're explicitly trying to determine Satisfiable/CounterSatisfiable *)
let nontheorem = ref false

(** proof output *)
type proofkind = TSTP | CoqScript | CoqSPfTerm | Dedukti | HOCore | Model | ModelTrue | IsarScript | PfInfo | PfUseful | PfFormdeps
let mkproofterm : proofkind option ref = ref None
let mkprooftermp () = OptionE.is_some !mkproofterm
let pfusefulout : string option ref = ref None
let pfformdepsout : string option ref = ref None
let proofoutfile : string option ref = ref None

(** 0 should mean print nothing (command line option -verb 0) *)
let verbosity : int ref = ref 1

(** testing & debugging *)
let printnumaxiomsflag : bool ref = ref false
let selectaxiomsflag : bool ref = ref false
let select_axioms_list : int list ref = ref [];;
let foffiles = ref false

(** TODO: is this really useful? *)
let recognize_hashroots = ref false



let read_proofkind = function
  "tstp" -> TSTP
| "coqscript" -> CoqScript
| "coqspfterm" -> CoqSPfTerm
| "dedukti" -> Dedukti
| "hocore" -> HOCore
| "model" -> Model
| "modeltrue" -> ModelTrue
| "isar" -> IsarScript
| "info" -> PfInfo
| "useful" -> PfUseful
| "formdeps" -> PfFormdeps
| p -> raise (InputHelpError ("Unknown kind of proof " ^ p ^ " for -p"))

let set_problemfile p =
  if !problemfile = "" then problemfile := p
  else raise (InputHelpError ("Multiple problem files passed: " ^ !problemfile
  ^ " and " ^ p))

let enslave args = slaveargs := List.rev_append args !slaveargs

let help_lines =
[ "Usage: satallax [OPTIONS]... <problemfile>"
; ""
; "Paths:"
; "  -M <dir> : Directory in which the mode/schedule files are stored"
; "  -P <file> : PicoMUS executable file"
; "  -E <file> : Eprover executable file"
; ""
; "Flag settings:"
; "  -m <mode> : Mode"
; "  -s <schedule> : Schedule of modes to try (previously called the 'strategy schedule')"
; "  -flag <name> <value> : Override a flag with a given value"
; ""
; "Termination criteria:"
; "  -t <timeout in seconds> : Time after which the prover terminates"
; "  -inferences <int> : Inferences after which the prover terminates"
; ""
; "Satisfiability:"
; "  -i : Do not report Satisfiable or CounterSatisfiable (problem is incomplete)"
; "  -N : Try to determine if the problem is Satisfiable or CounterSatisfiable"
; ""
; "Proof output:"
; "  -p <tstp|coqscript|coqspfterm|dedukti|hocore|modeltrue|model|isar|info|useful|formdeps> :"
; "    Output proof/proof object/proof info"
; "  -c <file> : Write proof to a file"
; "  -pfusefulout <file> : File in which to save info about what was useful in the proof"
; "  -pfformdepsout <file> : File in which to save what formulas/dependencies were in proof"
; ""
; "Verbosity:"
; "  -v : Verbose (equivalent to -verb 5)"
; "  -verb <int> : Verbosity of given level, -verb 0 means silent"
; "  -V : Print version number and quit"
]

let process_short_cmd_line_arg = function
  'v' -> enslave ["-v"]; verbosity := 5
| 'V' -> print_endline Version.version; exit 0
| 'h' -> enslave ["-h"]; recognize_hashroots := true
| 'i' -> enslave ["-i"]; completep := false
|  a  -> raise (InputHelpError ("Unknown command line argument " ^ String.make 1 a))

let process_cmd_line_arg = function
  "-m"::m::r -> mode := m :: !mode; r
| "-s"::s::r -> schedule_files := s :: !schedule_files; r
| "-M"::m::r -> Config.modedir := m; enslave ["-M"; m]; r
| "-P"::p::r -> Config.picomus := p; enslave ["-P"; p]; r
| "-E"::e::r -> Config.eprover := e; enslave ["-E"; e]; r
| "-t"::t::r -> timeout := Some (float_of_string t); r
| "-ht"::t::r -> hardtimeout := Some (float_of_string t); r
| "-c"::c::r ->
    proofoutfile := Some(c);
    enslave ["-c"; c]; r
| "-slave"::r -> slave := true; r
| "-preparsed"::r -> preparsed := true; r
| "-N"::r -> nontheorem := true; enslave ["-N"]; r
| "-flag"::f::v::r ->
    flag_overrides := (f, v)::!flag_overrides;
    enslave ["-flag"; f; v]; r
| "-p"::p::r ->
    mkproofterm := Some(read_proofkind p);
    enslave ["-p"; p]; r
| "-pfusefulout"::f::r -> pfusefulout := Some(f); enslave ["-pfusefulout";f]; r
| "-pfformdepsout"::f::r -> pfformdepsout := Some(f); enslave ["-pfformdepsout";f]; r
| "-verb"::v::r -> verbosity := int_of_string v; enslave ["-verb"; v]; r
| "-inferences"::i::r -> max_searchoptions := Some (int_of_string i); enslave ["-inferences"; i]; r
| "-numaxioms"::r -> printnumaxiomsflag := true; enslave ["-numaxioms"]; r
| "-selectaxioms"::n::r -> (*** This is only to experiment with different selections and (order) of the axioms/conjecture. ***)
    selectaxiomsflag := true;
    let num = int_of_string n in
    let axs = List.map int_of_string (ListE.take num r) in
    select_axioms_list := List.rev_append axs !select_axioms_list;
    ListE.drop num r
| "-foffiles"::r -> (*** This is only for testing and debugging interaction with FO provers like E. ***)
    foffiles := true; enslave ["-foffiles"]; r
| ""::r -> raise (InputHelpError "Problem processing command line arguments")
| "-"::r -> problemfile := ""; r
| opt::r ->
    if opt.[0] = '-'
    then String.iter process_short_cmd_line_arg (StringE.tail opt)
    else set_problemfile opt;
    r
| [] -> []

let process_command_line_args = function
  [] -> print_endline Version.version; List.iter print_endline help_lines; exit 1
| args -> ListE.iterate process_cmd_line_arg args
