open FormatE
open Fresh
open Hyps
open Problem
open Refutation
open Term
open TermP

(** put parentheses around a pretty printer and indent if lines are broken *)
let par f ff x = Format.fprintf ff "@[<1>(%a)@]" f x

let rec pp_dk_stp_parens p ff = function
  | Prop -> Format.fprintf ff "o"
  | Base(x) ->
      let x =
        try Hashtbl.find coq_names x
        with Not_found -> failwith ("dk_stp: " ^ x) in
      Format.fprintf ff "%s" x
  | Ar(a1,a2) -> if p then par pp_dk_ar ff (a1, a2) else pp_dk_ar ff (a1, a2)
and pp_dk_ar ff (a1, a2) =
  Format.fprintf ff "%a>%a"
    (pp_dk_stp_parens  true) a1
    (pp_dk_stp_parens false) a2

let pp_dk_stp = pp_dk_stp_parens true

let pp_dk_of_eta c (x, a) = Format.fprintf c "%s :@ eta@ %a" x pp_dk_stp a


let rec collect_lambdas bound l a m =
  let bound = Variables.push bound in
  let top = Variables.top bound in
  let l = (top, a) :: l in
  match m with
  | Lam (a, m) -> collect_lambdas bound l a m
  | _ -> (bound, List.rev l, m)


let pp_dk_poly c (op, a) = Format.fprintf c "%@%s %a" op pp_dk_stp a

let pp_dk_lam f g c (x, y) = Format.fprintf c "λ %a,@ %a" f x g y

let rec pp_dk_term_bound bound c m =
  match m with
  | Name(x,_) -> Format.fprintf c "%s" x
  | DB(i,a) -> Format.fprintf c "%s" (Variables.get i bound)
  | Imp   -> Format.fprintf c "imp"
  | False -> Format.fprintf c "false"
  | True  -> Format.fprintf c "true"
  | Neg   -> Format.fprintf c "not"
  | And   -> Format.fprintf c "and"
  | Or    -> Format.fprintf c "or"
  | Iff   -> Format.fprintf c "iff"
  | Forall(a) -> par pp_dk_poly c ("all"   , a)
  | Exists(a) -> par pp_dk_poly c ("exists", a)
  | Eq(a)     -> par pp_dk_poly c ("eq"    , a)
  | Choice(a) -> par pp_dk_poly c ("Eps"   , a)
  | Ap(Ap(Imp,m1),False) -> par (pp_dk_neg bound) c m1
  | Ap(Ap(And,m1),m2) -> par (pp_dk_bin bound) c (m1, "/\\", m2)
  | Ap(Ap( Or,m1),m2) -> par (pp_dk_bin bound) c (m1, "\\/", m2)
  | Ap(Ap(Iff,m1),m2) -> par (pp_dk_bin bound) c (m1, "<->", m2)
  | Ap(Ap(Imp,m1),m2) -> par (pp_dk_bin bound) c (m1, "-->", m2)
  | Ap(m1,m2) -> par (pp_dk_ap bound) c (m1, m2)
  | Lam(a,m) -> par (pp_dk_lambdas bound) c (a, m)

and pp_dk_neg bound c m = Format.fprintf c "~@ %a" (pp_dk_term_bound bound) m
and pp_dk_bin bound c (m1, op, m2) =
  Format.fprintf c "%a@ %s@ %a"
    (pp_dk_term_bound bound) m1 op
    (pp_dk_term_bound bound) m2
and pp_dk_ap_left bound c = function
  | Ap (m1, m2) -> pp_dk_ap bound c (m1, m2)
  | m -> pp_dk_term_bound bound c m
and pp_dk_ap bound c (m1, m2) =
  Format.fprintf c "%a@ %a" (pp_dk_ap_left bound) m1 (pp_dk_term_bound bound) m2
and pp_dk_lambdas bound c (a, m) =
  let (bound, l, m) = collect_lambdas bound [] a m in
  let pp_etas = Format.(pp_print_list ~pp_sep:pp_print_space (par pp_dk_of_eta)) in
  pp_dk_lam pp_etas (pp_dk_term_bound bound) c (l, m)

let pp_dk_term = pp_dk_term_bound (Variables.make ())

let pp_dk_of_eps c (x, m) = Format.fprintf c "%s :@ eps@ %a" x pp_dk_term m


let dk_sig_basetypes c x =
  Format.fprintf c "symbol const %s : Type\n" x

let dk_sig_const c (x, a) =
  if not (List.mem_assoc x !coqsig_def)
  then Format.fprintf c "symbol const %a\n" pp_dk_of_eta (x, a)

let dk_sig_def c (x, a) =
  try
    let m = List.assoc x !coqsig_def in
    Format.fprintf c "definition %a ≔@ %a\n" pp_dk_of_eta (x, a) pp_dk_term m
  with Not_found -> ()

let dk_sig_hyp c (x, t) =
  Format.fprintf c "symbol const %a\n" pp_dk_of_eps (x, coqnorm t)

let dk_sig_conjecture c (name, conj) =
  Format.fprintf c "definition %a" pp_dk_of_eps (name, coqnorm conj)

let get_conjecture = function
  | Some (t,_) -> (Hashtbl.find coq_hyp_names !conjecturename, t)
  | None -> ("claim", False)

let pp_dk_sig c () =
  let modules = ["logic"; "rules"; "identities"] in
  List.iter (Format.fprintf c "require open %s\n") modules;
  List.iter (dk_sig_basetypes c) !coqsig_base;
  List.iter (dk_sig_const c) !coqsig_const;
  List.iter (dk_sig_def c) !coqsig_const;
  List.iter (dk_sig_hyp c) !coqsig_hyp;
  Format.fprintf c "%a ≔@ " dk_sig_conjecture (get_conjecture !conjecture)

let dk_sig c = pp_to_out_channel c pp_dk_sig ()


type typ =
    Stp of stp
  | Trm of trm

let const x a = (Some x, Stp a)
let hyp m = (None, Trm m)

type context =
  { consts : (string * stp) list
  ;   hyps : (trm * string) list
  ; next_hyp : int
  }

let get_hyp rule tm ctx =
  lookup rule (coqnorm tm) ctx.hyps

let with_fresh ctx f =
  f { ctx with next_hyp = ctx.next_hyp + 1 } ctx.next_hyp

let rec with_name ctx id f =
  match id with
  | None ->
      with_fresh ctx
        (fun ctx fresh ->
          let name = "H" ^ string_of_int fresh in
          if Hashtbl.mem coq_used_names name
          then with_name ctx None f
          else f ctx name
        )
  | Some x -> f ctx x

let add_to_context ctx (x, t) =
  match t with
  | Stp a -> { ctx with consts = (x, a) :: ctx.consts }
  | Trm m -> { ctx with hyps = (coqnorm m, x) :: ctx.hyps }

let name_and_bind ctx (x, t) =
  with_name ctx x (fun ctx x -> (add_to_context ctx (x, t), x))

let name_binders ctx binders =
  List.fold_right
    (fun (x, t) (ctx, binders) ->
      let (ctx, x) = name_and_bind ctx (x, t) in
      (ctx, (x, t) :: binders))
    binders (ctx, [])

let pp_binder with_typ ff (x, t) =
  if with_typ then
    match t with
    | Stp a -> Format.fprintf ff "%a" (par pp_dk_of_eta) (x, a)
    | Trm m -> Format.fprintf ff "%a" (par pp_dk_of_eps) (x, m)
  else Format.pp_print_string ff x

let abstract binders ctx (pp : context -> 'a pp) ff (x : 'a) =
  let (ctx, binders) = name_binders ctx binders in
  let pp_binders = Format.(pp_print_list ~pp_sep:pp_print_space (pp_binder true)) in
  par (pp_dk_lam pp_binders (pp ctx)) ff (binders, x)

let inhabit (x, a) ctx pp ff y =
  Format.fprintf ff "Inh %a false %a"
    pp_dk_stp a (abstract [const x a] ctx pp) y

let inhabits m ctx (pp : context -> 'a pp) ff (y : 'a) =
  let xa = find_fresh_consts m ctx.consts in
  (* create a function of the shape:
    (fun ctx -> inhabit c ctx (x1, a1)
    (fun ctx -> inhabit c ctx (x2, a2)
    ...
    (fun ctx -> f ctx)))
  *)
  let pps = List.fold_right (fun (x, a) pp -> (fun ctx -> inhabit (x, a) ctx pp)) xa pp in
  pps ctx ff y

let eq_sym ff (a, s, t, h) =
  Format.fprintf ff "(%@eq_sym@ %a@ %a@ %a@ %s)"
    pp_dk_stp a pp_dk_term s pp_dk_term t h

let pp_dk_let m f1 f2 ctx ff (x1, x2) =
  let (ctx, id) = name_and_bind ctx (None, Trm m) in
  Format.fprintf ff "let %s ≔@ %a in@ %a"
    id f1 x1 (f2 ctx) x2

let eq_sym_maybe ctx ff (a, (h, sym)) =
  let h' = get_hyp "eq_sym_maybe" h ctx in
  match sym with
    Some (x, y) -> eq_sym ff (a, x, y, h')
  | None -> Format.fprintf ff "%s" h'

let conflict c hyp s2 ns2 =
  match (s2, ns2) with
  (* s2: s21 = s22
   * ns2: (ns22 = ns21) -> False
   *)
  | (Ap(Ap(Eq(a),s21),s22),Ap(Ap(Imp,Ap(Ap(Eq(_),ns22),ns21)),False))
    when s21 = ns21 && s22 = ns22 ->
    begin
      try
        (* find s22 = s21 in hyp *)
        let h1 = List.assoc (Ap(Ap(Eq(a),s22),s21)) hyp in
        Format.fprintf c "%s %s" (lookup "Conflict" ns2 hyp) h1
      with
      | Not_found ->
        begin
          try
            (* find (s21 = s22) -> False in hyp *)
            let h2 = List.assoc (Ap(Ap(Imp,(Ap(Ap(Eq(a),s21),s22))),False)) hyp in
            Format.fprintf c "%s %s" h2 (lookup "Conflict" s2 hyp)
          with
          | Not_found ->
              Format.fprintf c "%s %a"
                (lookup "Conflict" ns2 hyp)
                eq_sym (a, s21, s22, lookup "Conflict" s2 hyp)
        end
    end
  | _ -> Format.fprintf c "%s %s" (lookup "Conflict" ns2 hyp) (lookup "Conflict" s2 hyp)

let trans_order (w, z, h1) (v, u, h2) (s, t) =
  let wz = (h1, None)
  and zw = (h1, Some (w, z))
  and vu = (h2, None)
  and uv = (h2, Some (v, u)) in

  if coqnorm w = coqnorm s then
    begin
           if coqnorm v = coqnorm t then (z, wz, uv)
      else if coqnorm u = coqnorm t then (z, wz, vu)
      else failwith "Trans"
    end
  else if coqnorm z = coqnorm s then
    begin
           if coqnorm v = coqnorm t then (w, zw, uv)
      else if coqnorm u = coqnorm t then (w, zw, vu)
      else failwith "Trans"
    end
  else if coqnorm w = coqnorm t then
    begin
           if coqnorm v = coqnorm s then (z, uv, wz)
      else if coqnorm u = coqnorm s then (z, uv, zw)
      else failwith "Trans"
    end
  else if coqnorm z = coqnorm t then
    begin
           if coqnorm v = coqnorm s then (w, vu, wz)
      else if coqnorm u = coqnorm s then (w, vu, zw)
      else failwith "Trans"
    end
  else
    failwith "Trans"

let rewrite m =
  let ao a = Ar (a, Prop) in
  let aao a = Ar (a, ao a) in
  let ooo = aao Prop in
  let str_stp a = pp_to_string pp_dk_stp a in
  let ap_type x a = "(" ^ x ^ " " ^ str_stp a ^ ")" in

  let leibniz = function
    | (Ap(DB(0,_),DB(2,_)),
       Ap(DB(0,_),DB(1,_))) ->
        ("1", ao)
    | (Ap(Ap(Imp,(Ap(DB(0,_),DB(2,_)))),False),
       Ap(Ap(Imp,(Ap(DB(0,_),DB(1,_)))),False)) ->
        ("2", ao)
    | (Ap(Forall(_),Lam(_,(Ap(Ap(DB(1,_),DB(0,_)),DB(0,_))))),
       Ap(Ap(DB(0,_),DB(2,_)),DB(1,_))) ->
        ("3", aao)
    | (Ap(Ap(Imp,(Ap(Ap(DB(0,_),DB(2,_)),DB(1,_)))),False),
       Ap(Ap(Imp,(Ap(Forall(_),Lam(_,(Ap(Ap(DB(1,_),DB(0,_)),DB(0,_))))))),False)) ->
        ("4", aao)
    | _ -> failwith "unexpected rewrite step with Leibniz-like formula" in

  match m with
  | True -> (Prop, "true", "(~ false)", "eq_true")
  | And -> (ooo, "and", "and_imp", "eq_and_imp")
  | Or -> (ooo, "or", "or_imp", "eq_or_imp")
  | Iff -> (ooo, "iff", "eq", "eq_iff")
  | Lam(Ar(Prop,Prop) as oo,Ap(Ap(Imp,Ap(Ap(Imp,DB(0,Prop)),False)),False)) ->
      (oo, "neg_neg", "id", "eq_neg_neg_id")
  | Exists(a) ->
      (ao (ao a), "exists", "nforall", ap_type "eq_exists_nforall" a)
  | Lam(a,Lam(_,Ap(Forall(_),Lam(_,(Ap(Ap(Imp,prem),concl)))))) ->
      let (id, tyfun) = leibniz (prem, concl) in
      let lhs = ap_type ("leib" ^ id ^ "_lhs") (tyfun a) in
      let rule = ap_type ("eq_leib" ^ id) a in
      (aao a, lhs, "eq", rule)
  | Eq(_) -> failwith "unexpected rewrite step with Eq" (*** symmetry handled by known now ***)
  | pt -> failwith ("unknown rewrite step: " ^ trm_str pt)

let dk_itp_thm_str = let open Refut in function
    Eq_sym -> "@eq_sym"
  | Eq_trans -> "@eq_trans"
  | Eq_trans_sym_r -> "@eq_symtrans1"
  | Eq_trans_sym_lr -> "@eq_symtrans2"
  | Eq_ind -> "@eqE'"
  | Eq_ind_sym -> "@eqEr'"

let rec pp_dk_ref ctx ff r =
  match r with
  | Fal(_) -> Format.fprintf ff "%s" (get_hyp "Fal" False ctx)
  | Conflict(s,ns) -> conflict ff ctx.hyps (coqnorm s) (coqnorm ns)
  | NegRefl(Ap(Ap(Imp,Ap(Ap(Eq(a),s),_)),False) as h) ->
      Format.fprintf ff "TRef@ %a@ %a@ %s"
        pp_dk_stp a pp_dk_term s (get_hyp "NegRefl" h ctx)
  | DoubleNegation(h,s,r1) ->
      Format.fprintf ff "%s %a"
        (get_hyp "DoubleNegation" h ctx)
        (abstract [hyp s] ctx pp_dk_ref) r1
  | Implication(h,((Ap(Ap(Imp,s),False)) as s'),t,r1,r2) ->
      disjunction_like ff  "TImp" h s t s' t r1 r2 ctx
  | NegImplication(h,s,((Ap(Ap(Imp,t),False)) as t'),r1) ->
      conjunction_like ff "TNImp" h s t s t' r1 ctx
  | Conjunction(h,s,t,r1) ->
      conjunction_like ff "TAnd" h s t s t r1 ctx
  | NegDisjunction(h,((Ap(Ap(Imp,s),False)) as s'),
                     ((Ap(Ap(Imp,t),False)) as t'),r1) ->
      conjunction_like ff "TNOr" h s t s' t' r1 ctx
  | Disjunction(h,s,t,r1,r2) ->
      disjunction_like ff "TOr" h s t s t r1 r2 ctx
  | NegConjunction(h,((Ap(Ap(Imp,s),False)) as s'),
                     ((Ap(Ap(Imp,t),False)) as t'),r1,r2) ->
      disjunction_like ff "TNAnd" h s t s' t' r1 r2 ctx
  | All((Ap(Forall(_),m1) as h),s,r1,a,m,n) ->
      inhabits n ctx all_like ff ("TAll", a, m1, h, n, r1, s)
  | NegExist(((Ap(Ap(Imp,Ap(Exists(_),m1)),False)) as h),s,r1,a,m,n) ->
      inhabits n ctx all_like ff ("TNEx", a, m1, h, n, r1, s)
  | Exist(((Ap(Exists(_),m1)) as h),s,r1,a,m,x) ->
      exists_like ff "TEx" a m1 h x r1 s ctx
  | NegAll(((Ap(Ap(Imp,Ap(Forall(_),m1)),False)) as h),s,r1,a,m,x) ->
      exists_like ff "TNAll" a m1 h x r1 s ctx
  | Cut(s,r1,r2) ->
      inhabits s ctx
        (fun ctx ff () ->
          Format.fprintf ff "(%a@ %a)"
            (abstract [hyp (neg s)] ctx pp_dk_ref) r2
            (abstract [hyp      s ] ctx pp_dk_ref) r1
        )
        ff ()
  | Trans((Ap(Ap(Eq(a),w),z) as h1),
          (Ap(Ap(Eq(_),v),u) as h2),
          (Ap(Ap(Eq(_),s),t) as st), r1) ->
      let (x, eq1, eq2) = trans_order (w, z, h1) (v, u, h2) (s, t) in
      Format.fprintf ff "Ttrans@ %a@ %a@ %a@ %a@ %a@ %a@ %a"
        pp_dk_stp a pp_dk_term s pp_dk_term t pp_dk_term x
        (eq_sym_maybe ctx) (a, eq1) (eq_sym_maybe ctx) (a, eq2)
        (abstract [hyp st] ctx pp_dk_ref) r1
  | Delta(h,s,x,r1) ->
      let h1 = get_hyp "Delta" h ctx in
      pp_dk_ref (add_to_context ctx (h1, Trm s)) ff r1
  | ChoiceR(eps,pred,s,t,r1,r2) ->
      inhabits s ctx
        (fun ctx ff () ->
          match eps with
          | Choice(a) ->
              Format.fprintf ff "TEps@ %a@ %a@ %a@ %a"
                pp_dk_stp a pp_dk_term pred
                (abstract [hyp s] ctx pp_dk_ref) r1
                (abstract [hyp t] ctx pp_dk_ref) r2
          | Name(x,Ar(Ar(a,Prop),_)) -> failwith "CHOICE-TODO"
          | _ -> failwith "eps is not a valid epsilon"
        )
        ff ()
  | EqualProp(h,s,t,r1,r2) ->
      equal_prop_like ff "TBQ"   s t h s t r1 (neg s) (neg t) r2 ctx
  | Aequivalenz(h,s,t,r1,r2) ->
      equal_prop_like ff "TIff"  s t h s t r1 (neg s) (neg t) r2 ctx
  | NegEqualProp(h,s,t,r1,r2) ->
      equal_prop_like ff "TBE"   s t h s (neg t) r1 (neg s) t r2 ctx
  | NegAequivalenz(h,s,t,r1,r2) ->
      equal_prop_like ff "TNIff" s t h s (neg t) r1 (neg s) t r2 ctx
  | EqualFunc(((Ap(Ap(Eq(Ar(a,b)),s1),s2)) as h),s,r1) ->
      equal_func_like ff "TFQ" a b s1 s2 h s r1 ctx
  | NegEqualFunc(((Ap(Ap(Imp,Ap(Ap(Eq(Ar(a,b)),s1),s2)),False)) as h),s,r1) ->
      equal_func_like ff "TFE" a b s1 s2 h s r1 ctx
  | KnownResult(s,name,al,r1) ->
      let pp_stps = Format.(pp_print_list ~pp_sep:pp_print_space pp_dk_stp) in
      let pp_ap ff (name, al) = Format.fprintf ff "%s %a" (dk_itp_thm_str name) pp_stps al in
      pp_dk_let s pp_ap pp_dk_ref ctx ff ((name, al), r1)
  | Rewrite(prefix,pt,pt',r1) ->
      let h = coqnorm (Ap(prefix,pt )) in
      let s = coqnorm (Ap(prefix,pt')) in
      let h1 = get_hyp "Rewrite" h ctx in
      begin
        match pt with
        | Lam((Ar(a,b)),Lam(_,Ap(DB(1,_),DB(0,_)))) ->
            (*** Skip etas - Mar 2012 ***)
            pp_dk_ref { ctx with hyps = (s, h1) :: ctx.hyps } ff r1
        | _ ->
            let (a, t1, t2, rule) = rewrite pt in
            Format.fprintf ff "TRew@ %a@ %s@ %s@ %a@ %s@ %s@ %a"
              pp_dk_stp a t1 t2 pp_dk_term prefix rule h1
              (abstract [hyp s] ctx pp_dk_ref) r1
      end
  | NYI(h,s,r1) -> failwith "NYI step found in pp_dk_ref"
  | Timeout -> failwith "Timeout step found in pp_dk_ref"
  | Decomposition(((Ap(Ap(Imp,Ap(Ap(Eq(b),tl),tr)),False)) as h1), ss, rs) ->
      decompose ctx ff (h1, b, tl, tr, List.rev ss, List.rev rs)
  | Mating(h1, h2, ss, rs) ->
      mate ctx ff h1 h2 ss rs
  | Confront(h1,h2,su,tu,sv,tv,r1,r2) ->
      confront ctx ff h1 h2 su tu sv tv r1 r2
  | _ -> failwith "unknown refutation case in pp_dk_ref"

and all_like ctx c (rule, a, m1, h, n, r1, s) =
  Format.fprintf c "%s@ %a@ %a@ %s@ %a@ %a"
    rule pp_dk_stp a pp_dk_term m1 (get_hyp rule h ctx) pp_dk_term n
    (abstract [hyp s] ctx pp_dk_ref) r1

and exists_like c rule a m1 h x r1 s ctx =
  let x = Hashtbl.find coq_names x in
  Format.fprintf c "%s@ %a@ %a@ %s@ %a"
    rule pp_dk_stp a pp_dk_term m1 (get_hyp rule h ctx)
    (abstract [const x a; hyp s] ctx pp_dk_ref) r1

and conjunction_like c rule h s t s' t' r1 ctx =
  Format.fprintf c "%s@ %a@ %a@ %s@ %a"
    rule pp_dk_term s pp_dk_term t (get_hyp rule h ctx)
    (abstract [hyp s'; hyp t'] ctx pp_dk_ref) r1

and disjunction_like c rule h s t s' t' r1 r2 ctx =
  Format.fprintf c "%s@ %a@ %a@ %s@ %a@ %a"
    rule pp_dk_term s pp_dk_term t (get_hyp rule h ctx)
    (abstract [hyp s'] ctx pp_dk_ref) r1
    (abstract [hyp t'] ctx pp_dk_ref) r2

and equal_prop_like c rule s t h s1 t1 r1 s2 t2 r2 ctx =
  Format.fprintf c "%s@ %a@ %a@ %s@ %a@ %a"
    rule pp_dk_term s pp_dk_term t (get_hyp rule h ctx)
    (abstract [hyp s1; hyp t1] ctx pp_dk_ref) r1
    (abstract [hyp s2; hyp t2] ctx pp_dk_ref) r2

and equal_func_like c rule a b s1 s2 h s r1 ctx =
  Format.fprintf c "%s@ %a@ %a@ %a@ %a@ %s@ %a"
    rule pp_dk_stp a pp_dk_stp b
    pp_dk_term s1 pp_dk_term s2 (get_hyp rule h ctx)
    (abstract [hyp s] ctx pp_dk_ref) r1

and decompose ctx c (h1, b, tl, tr, ss, rs) =
  let h1n = get_hyp "Decomposition" h1 ctx in
  if coqnorm tl = coqnorm tr then
    Format.fprintf c "TRef@ %a@ %a@ %s" pp_dk_stp b pp_dk_term tl h1n
  else
    match (ss,rs,tl,tr) with
      ((((Ap(Ap(Imp,Ap(Ap(Eq(a),s11),s12)),False)) as s1)::sr),(r1::rr),Ap(tl1,_),Ap(tr1,_)) ->
        let h2 = Ap(Ap(Imp,Ap(Ap(Eq(Ar(a,b)),tl1),tr1)),False) in
        Format.fprintf c "TDec@ %a@ %a@ %a@ %a@ %a@ %a@ %s@ %a@ %a"
          pp_dk_stp a pp_dk_stp b
          pp_dk_term s11 pp_dk_term s12
          pp_dk_term tl1 pp_dk_term tr1
          h1n
          (abstract [hyp h2] ctx decompose) (h2, Ar(a,b), tl1, tr1, sr, rr)
          (abstract [hyp s1] ctx pp_dk_ref) r1
    | _ -> failwith "decomposition failed to render as Dedukti term"

and mate ctx ff h1 h2 ss rs =
  match (h1,h2,List.rev ss,List.rev rs) with (*** ss and rs **)
  | (Ap(Ap(Imp,Ap(h1p,h1s)),False), Ap(h2q,h2t),
     ((Ap(Ap(Imp,Ap(Ap(Eq(a),s11),s12)),False)) as s1)::sr, r1::rr)
    when h1s = s11 && h2t = s12 ->
      let h3 = Ap(Ap(Imp,Ap(Ap(Eq(Ar(a,Prop)),h1p),h2q)),False) in
      Format.fprintf ff "TMat@ %a@ %a@ %a@ %a@ %a@ %s@ %s@ %a@ %a"
        pp_dk_stp (tpof h1s)
        pp_dk_term h2t pp_dk_term h1s
        pp_dk_term h2q pp_dk_term h1p
        (get_hyp "Mating" h2 ctx) (get_hyp "Mating" h1 ctx)
        (** FIXME (was already in original code coqspfterm.ml) **)
        (abstract [hyp h3] ctx decompose) (h3, Ar(a,Prop), h2q, h1p, sr, rr)
        (abstract [hyp s1] ctx pp_dk_ref) r1
  | (Ap(Ap(Imp,Ap(h1p,h1s)),False), Ap(h2q,h2t),
     ((Ap(Ap(Imp,Ap(Ap(Eq(a),s11),s12)),False)) as s1)::sr, r1::rr) ->
      let h3 = Ap(Ap(Imp,Ap(Ap(Eq(Ar(a,Prop)),h1p),h2q)),False) in
      let h4 = Ap(Ap(Imp,Ap(Ap(Eq(Ar(a,Prop)),h2q),h1p)),False) in

      let eq_sym ff (a, s, t) =
        Format.fprintf ff "%@eq_sym@ %a@ %a@ %a"
          pp_dk_stp a pp_dk_term s pp_dk_term t in

      Format.fprintf ff "TMat@ %a@ %a@ %a@ %a@ %a@ %s@ %s@ %a@ %a"
        pp_dk_stp (tpof h1s)
        pp_dk_term h2t pp_dk_term h1s
        pp_dk_term h2q pp_dk_term h1p
        (get_hyp "Mating" h2 ctx) (get_hyp "Mating" h1 ctx)
        (abstract [hyp h4] ctx (pp_dk_let h3 eq_sym decompose))
        ((a, h1p, h2q), (h3, Ar(a,Prop), h2q, h1p, sr, rr))
        (abstract [hyp s1] ctx pp_dk_ref) r1
  | (Ap(h1p,h1s), Ap(Ap(Imp,Ap(h2q,h2t)),False),
     ((Ap(Ap(Imp,Ap(Ap(Eq(a),s11),s12)),False)) as s1)::sr, r1::rr) ->
      let h3 = Ap(Ap(Imp,Ap(Ap(Eq(Ar(a,Prop)),h1p),h2q)),False) in
      Format.fprintf ff "TMat@ %a@ %a@ %a@ %a@ %a@ %s@ %s@ %a@ %a"
        pp_dk_stp a
        pp_dk_term h1s pp_dk_term h2t
        pp_dk_term h1p pp_dk_term h2q
        (get_hyp "Mating" h1 ctx) (get_hyp "Mating" h2 ctx)
        (abstract [hyp h3] ctx decompose) (h3, Ar(a,Prop), h1p, h2q, sr, rr)
        (abstract [hyp s1] ctx pp_dk_ref) r1
  | _ -> failwith ("mating step did not match expected: " ^ trm_str h1 ^ " :: " ^ trm_str h2)

and confront ctx ff h1 h2 su tu sv tv r1 r2 =
  let sun = coqnorm su in
  let tvn = coqnorm tv in
  let is_eq = function Ap(Ap(Eq(_),_),_) -> true | _ -> false in
  (*** h1 and h2 may be reversed ***)
  let (h1, h2) = if is_eq (coqnorm h1) then (h1, h2) else (h2, h1) in

  (* TODO: is it really necessary to match on sun and not on su? *)
  match (sun,tvn) with
  | (Ap(Ap(Imp,Ap(Ap(Eq(a),s),u)),False),
     Ap(Ap(Imp,Ap(Ap(Eq(_),t),v)),False)) ->
      Format.fprintf ff "TCon@ %a@ %a@ %a@ %a@ %a@ %s@ %s@ %a@ %a"
        pp_dk_stp a
        pp_dk_term s pp_dk_term t
        pp_dk_term u pp_dk_term v
        (get_hyp "Confront" h1 ctx) (get_hyp "Confront" h2 ctx)
        (abstract [hyp su; hyp tu] ctx pp_dk_ref) r1
        (abstract [hyp sv; hyp tv] ctx pp_dk_ref) r2
  | _ -> failwith "confront does not match"


let init_context () =
  { hyps = List.fold_left (fun l (s,pt) -> (coqnorm pt,s)::l) [] !coqsig_hyp
  ; consts = !coqsig_const
  ; next_hyp = 0
  }

let get_conjecture () =
  match !conjecture with
    Some (con, _) ->
      begin
        let con = coqnorm con in
        match con with
        | False -> None
        | _ -> Some con
      end
  | None -> None

let pp_dk_ref_complete ff r =
  let ctx = init_context () in
  match get_conjecture () with
    Some con ->
      Format.fprintf ff "NNPP@ %a@ %a\n" pp_dk_term con
        (abstract [hyp (neg con)] ctx pp_dk_ref) r
  | None -> Format.fprintf ff "%a\n" (pp_dk_ref ctx) r

let ref_dedukti c = pp_to_out_channel c pp_dk_ref_complete
