open Fresh
open Hyps
open Problem
open Refutation
open Term
open TermP


let isarize_name x =
  if String.length x > 0 then
    if String.get x 0 = '_' then
      "eigen" ^ x
    else if String.get x 0 = '@' then
      String.sub x 1 ((String.length x) - 1)
    else
      if (String.length x > 5 && String.sub x 0 5 = "eigen") then
        "not" ^ x
      else
        (*Rename constants to avoid clashes with keywords*)
        let x =
          try Hashtbl.find coq_names x with
              Not_found -> x
        in x
  else failwith "Name cannot have zero length"


let rec pp_stp_parens p ff = function
  | Prop -> Format.fprintf ff "o"
  (* FIXME: this may clash with a base type that's really called "i" *)
  | Base("$i") -> Format.fprintf ff "i"
  | Base(x) -> Format.fprintf ff "%s" x
  | Ar(a1,a2) ->
      if p then Format.fprintf ff "(";
      pp_stp_parens true ff a1;
      Format.fprintf ff "=>";
      pp_stp_parens false ff a2;
      if p then Format.fprintf ff ")"

let pp_stp = pp_stp_parens true

let print_stp_isar c a p =
  FormatE.pp_to_out_channel c (pp_stp_parens p) a



let dest_lambda = function
  | Lam(a, m) -> Some(a, m)
  | _ -> None
let dest_forall = function
  | Ap(Forall(a), Lam(_, m)) -> Some(a, m)
  | _ -> None
let dest_exists = function
  | Ap(Exists(a), Lam(_, m)) -> Some(a, m)
  | _ -> None

let rec pp_term_bound bound c m =
  match m with
  | Name(x,_) -> Format.fprintf c "%s" (isarize_name x)
  | DB(i,a) -> Format.fprintf c "%s" (Variables.get i bound)
  | False -> Format.fprintf c "False"
  | True  -> Format.fprintf c "True"
  | Neg   -> Format.fprintf c "Not"
  | And   -> Format.fprintf c "(&)"
  | Or    -> Format.fprintf c "(|)"
  | Iff
  | Eq(_)     -> Format.fprintf c "(=)"
  | Choice(a) -> Format.fprintf c "Eps"
  | Forall(a) -> Format.fprintf c "All"
  | Exists(a) -> Format.fprintf c "Ex"
  | Ap(Forall(a),Lam(_,m)) -> Format.fprintf c "(∀%a)" (pp_forall bound) (a, m)
  | Ap(Exists(a),Lam(_,m)) -> Format.fprintf c "(∃%a)" (pp_exists bound) (a, m)
  | Ap(Choice(a),Lam(_,m)) -> Format.fprintf c "(ϵ%a)" (pp_choice bound) (a, m)
  | Ap(Ap(Imp,m1),False) -> Format.fprintf c "(~ %a)" (pp_term_bound bound) m1
  | Ap(Ap(Eq(_),m1),m2)
  | Ap(Ap(Iff,m1),m2) -> pp_bin bound c m1 " = " m2;
  | Ap(Ap(And,m1),m2) -> pp_bin bound c m1 " & " m2;
  | Ap(Ap( Or,m1),m2) -> pp_bin bound c m1 " | " m2;
  | Ap(Ap(Imp,m1),m2) -> pp_bin bound c m1 " --> " m2
  | Ap(Imp,m1) -> pp_term_bound bound c (Lam(Prop,Ap(Ap(Imp,shift m1 0 1),DB(0,Prop))))
  | Imp -> pp_term_bound bound c (Lam(Prop,Lam(Prop,Ap(Ap(Imp,DB(1,Prop)),DB(0,Prop)))))
  | Ap(m1,m2) -> pp_bin bound c m1 " " m2
  | Lam(a,m) -> Format.fprintf c "(λ%a)" (pp_lambda bound) (a, m)

and pp_bin bound c m1 op m2 =
  Format.fprintf c "(%a%s%a)" (pp_term_bound bound) m1 op (pp_term_bound bound) m2
and pp_binder terminal f bound c (a, m) =
  let bound = Variables.push bound in
  let top = Variables.top bound in
  Format.fprintf c " (%s::%a)" top pp_stp a;
  match f m with
  | Some (a, m) -> pp_binder terminal f bound c (a, m)
  | None -> Format.fprintf c "%s %a" terminal (pp_term_bound bound) m

and pp_lambda bound = pp_binder "." dest_lambda bound
and pp_forall bound = pp_binder "." dest_forall bound
and pp_exists bound = pp_binder "." dest_exists bound
and pp_choice bound = pp_binder "." (fun _ -> None) bound

let pp_term = pp_term_bound (Variables.make ())

let trm_to_isar c m bound =
  FormatE.pp_to_out_channel c (pp_term_bound bound) m


let sig_basetypes c x =
  if x <> "i" (*FIXME "$i"="i"*) then
    Format.fprintf c "typedecl %s\n" x

let sig_const c (x, a) =
  if not (List.mem_assoc x !coqsig_def)
  then Format.fprintf c "consts %s :: \"%a\"\n" x pp_stp a

let sig_def c (x, a) =
  try
    let m = List.assoc x !coqsig_def in
    Format.fprintf c "definition %s :: \"%a\"\n where \"%s == %a\"\n"
      x pp_stp a x pp_term m
  with Not_found -> ()

let sig_hyp c (x, t) =
  Format.fprintf c "assumes %s : \"%a\"\n" x pp_term (coqnorm t)

let sig_conjecture c (name, conj) =
  Format.fprintf c "shows %s : \"%a\"\n" name pp_term (coqnorm conj)

let get_conjecture = function
  | Some (t,_) -> (Hashtbl.find coq_hyp_names !conjecturename, t)
  | None -> ("claim", False)

let pp_strings = Format.(pp_print_list ~pp_sep:pp_print_space pp_print_string)

(** Determine the name of the Isabelle theory from
    the name of the proof script file. *)
let theory_name = function
  | Some filename -> Filename.(remove_extension (basename filename))
  | None -> "SatallaxProblem"

let pp_sig c filename =
  Format.fprintf c "theory %s\n" (theory_name filename);
  Format.fprintf c "imports Satallax\n";
  Format.fprintf c "begin\n";
  List.iter (sig_basetypes c) !coqsig_base;
  List.iter (sig_const c) !coqsig_const;
  List.iter (sig_def c) !coqsig_const;
  Format.fprintf c "\nlemma\n";
  List.iter (sig_hyp c) !coqsig_hyp;
  sig_conjecture c (get_conjecture !conjecture);
  (* FIXME: Currently all definitions are unfolded, irrespective of whether
     when they're used. This seems to reflect Satallax's usage anyway. *)
  if List.length !coqsig_def > 0 then
    let defs = List.map (fun (s, _) -> s ^ "_def") !coqsig_def in
    Format.fprintf c "unfolding %a\n" pp_strings defs

let isabellehol_sig c = FormatE.pp_to_out_channel c pp_sig


let next_fresh_name : int ref = ref 0
let get_fresh_name () =
  next_fresh_name := 1 + !next_fresh_name;
  "SOMENAME__" ^ (string_of_int (!next_fresh_name)) (*FIXME check for name collisions*)

let add_fresh_const c const n sp =
  let add x a =
    Printf.fprintf c "%sfixes (%s :: " sp x;
    print_stp_isar c a false;
    Printf.fprintf c ")\n";
  in add_fresh_consts add n const

let rec countup from for_many acc : int list =
  if for_many = 0 then List.rev acc
  else countup (from + 1) (for_many - 1) (from :: acc)

let isabelle_itp_thm_str = let open Refut in function
    Eq_sym -> "eq_sym"
  | Eq_trans -> "eq_trans"
  | Eq_trans_sym_r -> "eq_trans2"
  | Eq_trans_sym_lr -> "eq_trans3"
  | Eq_ind -> "eq_ind"
  | Eq_ind_sym -> "eq_ind_r"


(* copied from dedukti.ml *)
type context =
  { consts : (string * stp) list
  ;   hyps : (trm * string) list
  ; next_hyp : int
  }

let get_hyp rule tm ctx =
  lookup rule (coqnorm tm) ctx.hyps

let with_fresh ctx f =
  f { ctx with next_hyp = ctx.next_hyp + 1 } ctx.next_hyp


type rule_arg =
  | Rule_format
  | ROF of trm
  | RTHEN of rule
  | Rwhere of string * stp

and rule = string * rule_arg list

let rec pp_rule ctx ff = function
  | (rule, []) -> Format.fprintf ff "%s" rule
  | (rule, args) -> Format.fprintf ff "%s[%a]" rule (pp_rule_args ctx) args
and pp_rule_arg ctx ff = function
  | Rule_format -> Format.fprintf ff "rule_format"
  | ROF m -> Format.fprintf ff "OF %s" (get_hyp "" m ctx)
  | RTHEN rule -> Format.fprintf ff "THEN %a" (pp_rule ctx) rule
  | Rwhere (x, a) -> Format.fprintf ff "where %s = %a" x pp_stp a
and pp_rule_args ctx =
  let pp_comma ff () = Format.fprintf ff ",@ " in
  Format.(pp_print_list ~pp_sep:pp_comma (pp_rule_arg ctx))


let pp_ref ctx ff r =
  match r with
  | Fal(_) -> Format.fprintf ff "from %s show False by blast\n" (get_hyp "Fal" False ctx)
  | NegRefl(h) ->
      Format.fprintf ff "from %s have False by blast (*tab_refl*)\n" (get_hyp "NegRefl" h ctx);
      Format.fprintf ff "thus ?thesis by blast\n"





(*c is channel,
  hyp is shared hypotheses,
  h1 is new hypothesis' name,
  s and t are the new hypothesis terms,
  r1 and r2 are the remainder refutations*)
let rec ref_isabellehol1 c r hyp const sp=
  let sp' = sp ^ "  " in
  let sp'' = sp' ^ "  " in
  let tab_disj c hyp h1 s t r1 r2 =
    Printf.fprintf c "%sfrom %s have False\n" sp h1;
    Printf.fprintf c "%sproof\n" sp';
    Printf.fprintf c "%sassume %s : \"" sp'' h1;
    trm_to_isar c (coqnorm s) (Variables.make ());
    Printf.fprintf c "\"\n";
    ref_isabellehol1 c r1 ((coqnorm s, h1) :: hyp) const sp'';
    Printf.fprintf c "%snext\n" sp';
    Printf.fprintf c "%sassume %s : \"" sp'' h1;
    trm_to_isar c (coqnorm t) (Variables.make ());
    Printf.fprintf c "\"\n";
    ref_isabellehol1 c r2 ((coqnorm t, h1) :: hyp) const sp'';
    Printf.fprintf c "%sqed\n" sp';
    Printf.fprintf c "%sthus ?thesis by blast\n" sp' in
  (*like tab_disj, but with two hypotheses*)
  let tab_disj2 c hyp h1 h2 (s1, s2) (t1, t2) r1 r2 =
	  Printf.fprintf c "%sfrom %s have False\n" sp h1;
	  Printf.fprintf c "%sproof\n" sp;
	  Printf.fprintf c "%sassume %s : \"" sp' h1;
	  trm_to_isar c (coqnorm s1) (Variables.make ());
	  Printf.fprintf c "\"\n";
	  Printf.fprintf c "%s and %s : \"" sp' h2;
	  trm_to_isar c (coqnorm s2) (Variables.make ());
	  Printf.fprintf c "\"\n";
	  ref_isabellehol1 c r1 ((coqnorm s1, h1) :: (coqnorm s2, h2) :: hyp) const sp'';

	  Printf.fprintf c "%snext\n" sp';
	  Printf.fprintf c "%sassume %s : \"" sp' h1;
	  trm_to_isar c (coqnorm t1) (Variables.make ());
	  Printf.fprintf c "\"\n";
	  Printf.fprintf c "%s and %s : \"" sp' h2;
	  trm_to_isar c (coqnorm t2) (Variables.make ());
	  Printf.fprintf c "\"\n";
	  ref_isabellehol1 c r2 ((coqnorm t1, h1) :: (coqnorm t2, h2) :: hyp) const sp'';
    Printf.fprintf c "%sqed\n" sp';
    Printf.fprintf c "%sthus ?thesis by blast\n" sp'
  in
	match r with
    | Conflict(s,ns) ->
	      Printf.fprintf c "%sfrom %s %s show ?thesis by blast\n" sp (lookup "0" (coqnorm s) hyp) (lookup "1" (coqnorm ns) hyp)
    | Fal(_) ->
        Printf.fprintf c "%sfrom %s show False by blast\n" sp (lookup "2" False hyp);
    | NegRefl(s) ->
	      Printf.fprintf c "%sfrom %s have False by blast (*tab_refl*)\n" sp (lookup "3" (coqnorm s) hyp);
	      Printf.fprintf c "%sthus ?thesis by blast\n" sp
    | Implication(h,s,t,r1,r2) ->
	      let h1 = get_hyp_name() in
	        Printf.fprintf c "%snote %s = %s[THEN TImp[THEN mp]](*tab_imp*)\n" sp h1 (lookup "4" (coqnorm h) hyp);
          tab_disj c hyp h1 s t r1 r2
    | Disjunction(h,s,t,r1,r2) ->
	      let h1 = get_hyp_name() in
	        Printf.fprintf c "%snote %s = %s(*tab_or*)\n" sp h1 (lookup "5" (coqnorm h) hyp);
          tab_disj c hyp h1 s t r1 r2
    | NegConjunction(h,s,t,r1,r2) ->
	      let h1 = get_hyp_name() in
	        Printf.fprintf c "%snote %s = TNAnd[rule_format, OF %s](*tab_nand*)\n" sp h1 (lookup "6" (coqnorm h) hyp);
          tab_disj c hyp h1 s t r1 r2
    | NegImplication(h,s,t,r1) ->
	      let h1 = get_hyp_name() in
	      let h2 = get_hyp_name() in
	        Printf.fprintf c "%snote %s = %s[THEN TNegImp1]\n" sp h1 (lookup "7" (coqnorm h) hyp);
	        Printf.fprintf c "%snote %s = %s[THEN TNegImp2]\n" sp h2 (lookup "7" (coqnorm h) hyp);
	        ref_isabellehol1 c r1 ((coqnorm s, h1) :: (coqnorm t, h2) :: hyp) const sp
    | Conjunction(h,s,t,r1) ->
	      let h1 = get_hyp_name() in
	      let h2 = get_hyp_name() in
	        Printf.fprintf c "%snote %s = %s[THEN TAnd1]\n" sp h1 (lookup "8" (coqnorm h) hyp);
	        Printf.fprintf c "%snote %s = %s[THEN TAnd2]\n" sp h2 (lookup "8" (coqnorm h) hyp);
	        ref_isabellehol1 c r1 ((coqnorm s, h1) :: (coqnorm t, h2) :: hyp) const sp
    | NegDisjunction(h,s,t,r1) ->
	      let h1 = get_hyp_name() in
	      let h2 = get_hyp_name() in
	        Printf.fprintf c "%snote %s = TNor1[rule_format, OF %s]\n" sp h1 (lookup "9" (coqnorm h) hyp);
	        Printf.fprintf c "%snote %s = TNor2[rule_format, OF %s]\n" sp h2 (lookup "9" (coqnorm h) hyp);
	        ref_isabellehol1 c r1 ((coqnorm s, h1) :: (coqnorm t, h2) :: hyp) const sp
    | All(h,s,r1,a,m,n) ->
	      let const = add_fresh_const c const n sp in
	      let h1 = get_hyp_name() in
          (*Translated inference should look something like this:
            note H17 = H16[THEN spec, of "eigen__0 (eigen__2 eigen__5 eigen__6)"] *)
	        Printf.fprintf c "%snote %s = %s[THEN spec, of \"" sp h1 (lookup "10" (coqnorm h) hyp);
	        trm_to_isar c n (Variables.make ());
	        Printf.fprintf c "\"]\n";
	        ref_isabellehol1 c r1 ((coqnorm s, h1) :: hyp) const sp
    | NegAll(h,s,r1,a,m,x) ->
	      let h1 = get_hyp_name() in
	      let x = Hashtbl.find coq_names x in
	        Printf.fprintf c "%sfrom %s obtain eigen%s where %s : \"" sp (lookup "11" (coqnorm h) hyp) x h1;
	        trm_to_isar c (coqnorm s) (Variables.make ());
	        Printf.fprintf c "\" by blast\n";
	        ref_isabellehol1 c r1 ((coqnorm s, h1) :: hyp) ((x, a) :: const) sp
    | Exist(h,s,r1,a,m,x) ->
	      let h1 = get_hyp_name() in
	      let x = ( Hashtbl.find coq_names x ) in
          (*Translated inferences should look something like this:
            from H4 obtain eigen__1 where H5 : "rel_d eigen__0 eigen__1" by (erule TEx)   *)

	        Printf.fprintf c "%sfrom %s obtain eigen%s where %s : \"" sp (lookup "12" (coqnorm h) hyp) x h1;
	        trm_to_isar c (coqnorm s) (Variables.make ());
	        Printf.fprintf c "\" by (erule TEx)\n";
	        ref_isabellehol1 c r1 ((coqnorm s, h1) :: hyp) ((x, a) :: const) sp
    | NegExist(h,s,r1,a,m,n) ->
	      let const = add_fresh_const c const n sp in
	      let h1 = get_hyp_name() in
	        Printf.fprintf c "%snote %s = TNegEx[OF %s, where y = \"" sp h1 (lookup "13" (coqnorm h) hyp);
	        trm_to_isar c (coqnorm n) (Variables.make ());
	        Printf.fprintf c "\"]\n";
	        ref_isabellehol1 c r1 ((coqnorm s, h1) :: hyp) const sp
    | Mating(h1,h2, ss, rs) ->
        (*FIXME in the Coq code for tab_mat we also try swapping H1 and H2. Currently we don't emulate that here.*)
        assert ( ((neg_p h1) || (neg_p h2)) && not ((neg_p h1) && (neg_p h2)));
	      let h3 = get_hyp_name() in
          (*
            This rule seems to combine resolution with decomposition. It takes two facts "p s1 sn" and "~ q t1 tn"
            then we obtain a refutation by showing that "p = q" and "si = ti" for all i.
            Satallax's Coq reconstruction reduces this to decomposition (see the ltac definition for tab_mat) but
            we treat the whole mating here (mainly because we cannot rely on the compositionality of Satallax's approach
            in Coq for the time being).
          *)

        let (neg_hyp, pos_hyp) =
          if neg_p h1 then (h1, h2) else (h2, h1) in

    (*NOTE mostly duplicated from tab_dec*)
        let card = List.length ss in
        let proof_step_str =
          if card > 1 then
            "proof"
          else
            "proof -" in
        let fresh_fact_name = get_fresh_name () in
        let head = (*NOTE this was changed from tab_dec*)
          match neg_body neg_hyp(*this was h1 in tab_dec*) with
              Some h1' -> fst (bounded_head_spine card h1')
            | _ -> failwith "Could not determine head expression during Isar translation.2" in
        let indices = countup 0 card [] in
        let (custom_dec_prefix, custom_dec_suffix) =
          let (prefix, diseqs) =
            List.split (List.map (fun i ->
              ("s" ^ string_of_int i ^ " t" ^ string_of_int i,
               "(s" ^ string_of_int i ^ " ~= t" ^ string_of_int i ^ ")"))
              indices)
          in
            ("have " ^ fresh_fact_name ^ " : \"!! " ^ String.concat " " prefix ^ ". [|",
             "|] ==> " ^ String.concat " | " diseqs ^ "\" by blast") in
        (*FIXME this bit would be less of a mess if we use sprintf instead of printf*)
        let print_diseq () = (*NOTE this function was changed from that in the tab_dec handler*)
          let side v = " " ^ String.concat " " (List.map (fun i -> v ^ string_of_int i) indices)
          in
            trm_to_isar c head (Variables.make ());
	          Printf.fprintf c "%s" (side "s");
	          Printf.fprintf c "; ~ "; (*NOTE interesting that in Leo2 resolution covers the functionality of mating*)
            trm_to_isar c head (Variables.make ());
	          Printf.fprintf c "%s" (side "t");
        in
	        Printf.fprintf c "%s%s" sp custom_dec_prefix;
          print_diseq ();
	        Printf.fprintf c "%s\n" custom_dec_suffix;
	        Printf.fprintf c "%snote %s = %s[OF %s, OF %s]\n" sp h3 fresh_fact_name (lookup "14" (coqnorm pos_hyp) hyp) (lookup "15" (coqnorm neg_hyp) hyp); (*NOTE this line of the tab_dec handler was changed*)
	        Printf.fprintf c "%sfrom %s have False\n" sp h3;
          Printf.fprintf c "%s%s\n" sp' proof_step_str;

	        ignore(List.fold_right
            (fun (s, r) remaining ->
               Printf.fprintf c "%sassume %s : \"" sp'' h3;
	             trm_to_isar c (coqnorm s) (Variables.make ());
               Printf.fprintf c "\"\n";
               ref_isabellehol1 c r ((coqnorm s,h3)::hyp) const sp'';

               if remaining > 1 then Printf.fprintf c "%snext\n" sp';

               remaining - 1)
            (List.combine ss rs)
            card);

          Printf.fprintf c "%sqed\n" sp';
          Printf.fprintf c "%sthus ?thesis by blast\n" sp';
    | Decomposition(h1, ss, rs) ->
	      let h3 = get_hyp_name() in
        let card = List.length ss in
        let proof_step_str =
          if card > 1 then
            "proof"
          else
            "proof -" in
        let fresh_fact_name = get_fresh_name () in
        let head =
          match neg_body h1 with
            Some h1' ->
              begin
                match eq_body h1' with
                    Some (_, m, _) -> fst (bounded_head_spine card m)
                  | _ -> failwith "Could not determine head expression during Isar translation."
              end
            | _ -> failwith "Could not determine head expression during Isar translation." in
        let indices = countup 0 card [] in
        let (custom_dec_prefix, custom_dec_suffix) =
          let (prefix, diseqs) =
            List.split (List.map (fun i ->
              ("s" ^ string_of_int i ^ " t" ^ string_of_int i,
               "(s" ^ string_of_int i ^ " ~= t" ^ string_of_int i ^ ")"))
              indices)
          in
            ("have " ^ fresh_fact_name ^ " : \"!! " ^ String.concat " " prefix ^ ". ",
             " ==> " ^ String.concat " | " diseqs ^ "\" by blast") in
        (*FIXME this bit would be less of a mess if we use sprintf instead of printf*)
        let print_diseq () =
          let side v = " " ^ String.concat " " (List.map (fun i -> v ^ string_of_int i) indices)
          in
            trm_to_isar c head (Variables.make ());
	          Printf.fprintf c "%s" (side "s");
	          Printf.fprintf c " ~= ";
            trm_to_isar c head (Variables.make ());
	          Printf.fprintf c "%s" (side "t");
        in

	        Printf.fprintf c "%s%s" sp custom_dec_prefix;
          print_diseq ();
	        Printf.fprintf c "%s\n" custom_dec_suffix;
	        Printf.fprintf c "%snote %s = %s[OF %s]\n" sp h3 fresh_fact_name (lookup "16" (coqnorm h1) hyp);
	        Printf.fprintf c "%sfrom %s have False\n" sp h3;
          Printf.fprintf c "%s%s\n" sp' proof_step_str;

	        ignore(List.fold_right
            (fun (s, r) remaining ->
               Printf.fprintf c "%sassume %s : \"" sp'' h3;
	             trm_to_isar c (coqnorm s) (Variables.make ());
               Printf.fprintf c "\"\n";
               ref_isabellehol1 c r ((coqnorm s,h3)::hyp) const sp'';

               if remaining > 1 then Printf.fprintf c "%snext\n" sp';

               remaining - 1)
            (List.combine ss rs)
            card);

          Printf.fprintf c "%sqed\n" sp';
          Printf.fprintf c "%sthus ?thesis by blast\n" sp';
    | Confront(h1,h2,su,tu,sv,tv,r1,r2) ->
	      let h3 = get_hyp_name() in
	      let h4 = get_hyp_name() in
        let fresh_fact_name = get_fresh_name () in
	        Printf.fprintf c "%snote %s = TCON[OF %s, OF %s](*FIXME should also try swapping the previous OFs*)(*tab_con*)\n" sp fresh_fact_name (lookup "17" (coqnorm h1) hyp) (lookup "18" (coqnorm h2) hyp);

          (*FIXME this next bit is dirty -- might be better to adapt tab_disj2*)
          Printf.fprintf c "%sfrom %s have False\n" sp fresh_fact_name;
          Printf.fprintf c "%sproof\n" sp;
          Printf.fprintf c "%sassume %s : \"" sp' fresh_fact_name;
          Printf.fprintf c "(";
          trm_to_isar c (coqnorm su) (Variables.make ());
          Printf.fprintf c ") & (";
          trm_to_isar c (coqnorm tu) (Variables.make ());
          Printf.fprintf c ")";
          Printf.fprintf c "\"\n";
          Printf.fprintf c "%sfrom %s have %s : \"" sp' fresh_fact_name h3;
          trm_to_isar c (coqnorm su) (Variables.make ());
          Printf.fprintf c "\" by blast\n";
          Printf.fprintf c "%sfrom %s have %s : \"" sp' fresh_fact_name h4;
          trm_to_isar c (coqnorm tu) (Variables.make ());
          Printf.fprintf c "\" by blast\n";
          ref_isabellehol1 c r1 ((coqnorm su, h3) :: (coqnorm tu, h4) :: hyp) const sp'';

          Printf.fprintf c "%snext\n" sp';
          Printf.fprintf c "%sassume %s : \"" sp' fresh_fact_name;
          Printf.fprintf c "(";
          trm_to_isar c (coqnorm sv) (Variables.make ());
          Printf.fprintf c ") & (";
          trm_to_isar c (coqnorm tv) (Variables.make ());
          Printf.fprintf c ")";
          Printf.fprintf c "\"\n";
          Printf.fprintf c "%sfrom %s have %s : \"" sp' fresh_fact_name h3;
          trm_to_isar c (coqnorm sv) (Variables.make ());
          Printf.fprintf c "\" by blast\n";
          Printf.fprintf c "%sfrom %s have %s : \"" sp' fresh_fact_name h4;
          trm_to_isar c (coqnorm tv) (Variables.make ());
          Printf.fprintf c "\" by blast\n";
          ref_isabellehol1 c r2 ((coqnorm sv, h3) :: (coqnorm tv, h4) :: hyp) const sp'';
          Printf.fprintf c "%sqed\n" sp';
          Printf.fprintf c "%sthus ?thesis by blast\n" sp'
    | Trans(h1,h2,su,r1) ->
	      let h3 = get_hyp_name() in
	        Printf.fprintf c "%s(*FIXME only one out of the next four lines should be used --- comment the others out.*)\n" sp;
	        Printf.fprintf c "%snote %s = Ttrans[rule_format, OF %s, OF %s]\n" sp h3 (lookup "19" (coqnorm h1) hyp) (lookup "20" (coqnorm h2) hyp);
	        Printf.fprintf c "%snote %s = Ttrans[rule_format, OF %s[symmetric], OF %s]\n" sp h3 (lookup "19" (coqnorm h1) hyp) (lookup "20" (coqnorm h2) hyp);
	        Printf.fprintf c "%snote %s = Ttrans[rule_format, OF %s, OF %s[symmetric]]\n" sp h3 (lookup "19" (coqnorm h1) hyp) (lookup "20" (coqnorm h2) hyp);
	        Printf.fprintf c "%snote %s = Ttrans[rule_format, OF %s[symmetric], OF %s[symmetric]]\n" sp h3 (lookup "19" (coqnorm h1) hyp) (lookup "20" (coqnorm h2) hyp);
	        ref_isabellehol1 c r1 ((coqnorm su,h3)::hyp) const (sp^" ");
    | NegEqualProp(h,s,t,r1,r2) ->
	      let h1 = get_hyp_name() in
	      let h2 = get_hyp_name() in
	        Printf.fprintf c "%snote %s = TBE[rule_format, OF %s](*tab_be*)\n" sp h1 (lookup "21" (coqnorm h) hyp);
          tab_disj2 c hyp h1 h2 (s, neg t) (neg s, t) r1 r2
    | EqualProp(h,s,t,r1,r2) ->
	      let h1 = get_hyp_name() in
	      let h2 = get_hyp_name() in
	        Printf.fprintf c "%snote %s = TBQ[rule_format, OF %s](*tab_bq*)\n" sp h1 (lookup "22" (coqnorm h) hyp);
          tab_disj2 c hyp h1 h2 (s, t) (neg s, neg t) r1 r2
    | NegAequivalenz(h,s,t,r1,r2) ->
	      let h1 = get_hyp_name() in
	      let h2 = get_hyp_name() in
	        Printf.fprintf c "%snote %s = TNIff[rule_format, OF %s](*tab_negiff*)\n" sp h1 (lookup "23" (coqnorm h) hyp);
          tab_disj2 c hyp h1 h2 (s, neg t) (neg s, t) r1 r2
    | Aequivalenz(h,s,t,r1,r2) ->
	      let h1 = get_hyp_name() in
	      let h2 = get_hyp_name() in
	        Printf.fprintf c "%snote %s = TIff[rule_format, OF %s](*tab_iff*)\n" sp h1 (lookup "24" (coqnorm h) hyp);
          tab_disj2 c hyp h1 h2 (s, t) (neg s, neg t) r1 r2
    | NegEqualFunc(h,s,r1) ->
	      let h1 = get_hyp_name() in
	        Printf.fprintf c "%snote %s = TFE[THEN mp, OF %s]\n" sp h1 (lookup "25" (coqnorm h) hyp);
	        ref_isabellehol1 c r1 ((coqnorm s, h1) :: hyp) const sp
    | EqualFunc(h,s,r1) ->
	      let h1 = get_hyp_name() in
	        Printf.fprintf c "%snote %s = TFQ[THEN mp, OF %s]\n" sp h1 (lookup "26" (coqnorm h) hyp);
	        ref_isabellehol1 c r1 ((coqnorm s, h1) :: hyp) const  sp
    | ChoiceR(eps,pred,s,t,r1,r2) ->
        let const = add_fresh_const c const pred sp in
        let h1 = get_hyp_name() in
          begin
            match eps with
              | Choice(a) ->
                  (*Translated inferences should look something like this:
                    let ?p = "% (X1 :: i). f X1 = eigen__11"
                    have H4 : "(! x. ~ ?p x)" by (rule TSeps[where 'A = "i" and p = ?p, THEN mp], rule impI, insert H2, blast)
                  *)
                  let termname = "?" ^ get_fresh_name ()
                  in
	                  Printf.fprintf c "%slet %s = \"" sp termname;
	                  trm_to_isar c pred (Variables.make ());
	                  Printf.fprintf c "\"\n";
	                  Printf.fprintf c "%shave %s : \"! x. ~ %s x\" " sp h1 termname;
	                  Printf.fprintf c "by (rule TSeps[where 'A = \"";
	                  print_stp_isar c a true;
                    Printf.fprintf c "\" and p = %s, THEN mp], rule impI, insert %s, blast)\n" termname (String.concat " " (List.map snd hyp));
	                  ref_isabellehol1 c r1 ((coqnorm s,h1)::hyp) const (sp ^ " ");
              | Name(x,Ar(Ar(a,Prop),_)) ->
                  (*FIXME this rule is broken, e.g., try reconstructing proof from
                    running ./bin/satallax -m mode238 -p isar /home/nik/TPTP-v5.5.0/Problems/SYO/SYO538^1.p*)

                  (*NOTE this is largely adapted from the handler of tab_choice*)
                  let termname = "?" ^ get_fresh_name ()
                  in
	                  Printf.fprintf c "%slet %s = \"" sp termname;
	                  trm_to_isar c pred (Variables.make ());
	                  Printf.fprintf c "\"\n";
	                  Printf.fprintf c "%shave %s : \"! x. ~ %s x\" " sp h1 termname;

	                  Printf.fprintf c "\n(*FIXME only one from the next three should be used --- comment out the others.*)\n";
	                  Printf.fprintf c "by (rule TSeps'[where 'A = \"";
	                  print_stp_isar c a true;
                    Printf.fprintf c "\", THEN spec, of \"";
	                  trm_to_isar c eps (Variables.make ());
                    Printf.fprintf c "\", THEN spec, of \"%s\", THEN mp, OF %s, THEN mp], rule impI, insert %s, simp)\n" termname (Choice.get_Choicop_axiom x a hyp) (String.concat " " (List.map snd hyp));

                    (*FIXME DRY principle -- this is adapted from above. The only change is "TSeps'" to "TSeps''"*)
	                  Printf.fprintf c "\n";
	                  Printf.fprintf c "by (rule TSeps''[where 'A = \"";
	                  print_stp_isar c a true;
                    Printf.fprintf c "\", THEN spec, of \"";
	                  trm_to_isar c eps (Variables.make ());
                    Printf.fprintf c "\", THEN spec, of \"%s\", THEN mp, OF %s, THEN mp], rule impI, insert %s, simp)\n" termname (Choice.get_Choicop_axiom x a hyp) (String.concat " " (List.map snd hyp));

                    (*FIXME DRY principle -- this is adapted from above. The only change is "TSeps'" to "TSeps'''"*)
	                  Printf.fprintf c "\n";
	                  Printf.fprintf c "by (rule TSeps'''[where 'A = \"";
	                  print_stp_isar c a true;
                    Printf.fprintf c "\", THEN spec, of \"";
	                  trm_to_isar c eps (Variables.make ());
                    Printf.fprintf c "\", THEN spec, of \"%s\", THEN mp, OF %s, THEN mp], rule impI, insert %s, simp)\n" termname (Choice.get_Choicop_axiom x a hyp) (String.concat " " (List.map snd hyp));

	                  ref_isabellehol1 c r1 ((coqnorm s,h1)::hyp) const (sp^" ");
	                  ref_isabellehol1 c r2 ((coqnorm t,h1)::hyp) const (sp^" ");
              | _ -> failwith "eps is not a valid epsilon"
          end
    | Cut(s,r1,r2) ->
	      let const = add_fresh_const c const s sp in
	      let h1 = get_hyp_name() in
        let termname = "?" ^ get_fresh_name ()
        in
	        Printf.fprintf c "%slet %s = \"" sp termname;
	        trm_to_isar c s (Variables.make ());
	        Printf.fprintf c "\"\n";
	        Printf.fprintf c "%shave %s : \"~ %s | %s\" by blast (*tab_cut*)\n" sp h1 termname termname;
          tab_disj c (((coqnorm (disj (neg s) s)), h1) :: hyp) h1 (neg s) s r2 r1
    | DoubleNegation(h,s,r1) ->
	      let h1 = get_hyp_name() in
	        Printf.fprintf c "%snote %s = notnotD[OF %s]\n" sp h1 (lookup "27" (coqnorm h) hyp);
	        ref_isabellehol1 c r1 ((coqnorm s,h1)::hyp) const sp;
    | Rewrite(prefix,pt,pt',r1) ->
	      let h =  coqnorm (Ap(prefix,pt)) in
	      let h1 = lookup "28" h hyp in
	      let s =  coqnorm (Ap(prefix,pt')) in
	      let h2 = get_hyp_name() in
	        begin
	          match pt with
		          | True ->
                  Printf.fprintf c "%snote %s = eq_ind[THEN spec, of \"True\", THEN spec, of \"" sp h2;
				          trm_to_isar c prefix (Variables.make ());
                  Printf.fprintf c "\", THEN mp, OF %s, THEN spec, of \"~ False\", THEN mp, OF eq_true]\n" h1;
		          | And ->
                  Printf.fprintf c "%snote %s = eq_ind[THEN spec, of \"(&)\", THEN spec, of \"" sp h2;
				          trm_to_isar c prefix (Variables.make ());
                  Printf.fprintf c "\", THEN mp, OF %s, THEN spec, of \"%% x y. ~ (x --> ~ y)\", THEN mp, OF eq_and_imp]\n" h1;
		          | Or ->
                  Printf.fprintf c "%snote %s = eq_ind[THEN spec, of \"(|)\", THEN spec, of \"" sp h2;
				          trm_to_isar c prefix (Variables.make ());
                  Printf.fprintf c "\", THEN mp, OF %s, THEN spec, of \"%% x y. ((~ x) --> y)\", THEN mp, OF eq_or_imp]\n" h1;
		          | Iff ->
                  Printf.fprintf c "%snote %s = eq_ind[THEN spec, of \"(=)\", THEN spec, of \"" sp h2;
				          trm_to_isar c prefix (Variables.make ());
                  Printf.fprintf c "\", THEN mp, OF %s, THEN spec, of \"(=)\", THEN mp, OF eq_iff]\n" h1;
		          | Exists(_) ->
                  (*Translated inferences should look something like this:
                    note H2 = eq_ind[THEN spec, of "% p. ? x. p x", THEN spec, of "% (X1::(bool=>bool)=>bool). ~ X1 (% X2::bool. X2)", THEN mp, OF H0, THEN spec, of "% p. ~ (! x. ~ p x)", THEN mp, OF eq_exists_nforall] *)

                  (*FIXME horrible hardcoding*)
                    Printf.fprintf c "%snote %s = eq_ind[THEN spec, of \"%% p. ? x. p x\", THEN spec, of \"" sp h2;
				            trm_to_isar c prefix (Variables.make ());
                    Printf.fprintf c "\", THEN mp, OF %s, THEN spec, of \"%% p. ~ (! x. ~ p x)\", THEN mp, OF eq_exists_nforall]\n" h1;
		          | Eq(_) ->
                  Printf.fprintf c "%snote %s = eq_ind[THEN spec, of \"(=)\", THEN spec, of \"" sp h2;
				          trm_to_isar c prefix (Variables.make ());
                  Printf.fprintf c "\", THEN mp, OF %s, THEN spec, of \"%% s t. t = s\", THEN mp, OF eq_sym_eq]\n" h1;
		          | Lam(_,Lam(_,Ap(DB(1,_),DB(0,_)))) ->
                  Printf.fprintf c "%snote %s = eq_ind[THEN spec, of \"%% f x. f x\", THEN spec, of \"" sp h2;
				          trm_to_isar c prefix (Variables.make ());
                  Printf.fprintf c "\", THEN mp, OF %s, THEN spec, of \"%% f. f\", THEN mp, OF eq_eta]\n" h1;
		          | Lam(Ar(Prop,Prop),Ap(Ap(Imp,Ap(Ap(Imp,DB(0,Prop)),False)),False)) ->
                  Printf.fprintf c "%snote %s = eq_ind[THEN spec, of \"%% x. ~ ~ x\", THEN spec, of \"" sp h2;
				          trm_to_isar c prefix (Variables.make ());
                  Printf.fprintf c "\", THEN mp, OF %s, THEN spec, of \"%% x . x\", THEN mp, OF eq_neg_neg_id]\n" h1;
		          | Lam(_,Lam(_,Ap(Forall(_),Lam(_,(Ap(Ap(Imp,(Ap(DB(0,_),DB(2,_)))),(Ap(DB(0,_),DB(1,_)))) ))) )) ->
                  Printf.fprintf c "%snote %s = eq_ind[THEN spec, of \"%% s t. ! p. p s --> p t\", THEN spec, of \"" sp h2;
				          trm_to_isar c prefix (Variables.make ());
                  Printf.fprintf c "\", THEN mp, OF %s, THEN spec, of \"(=)\", THEN mp, OF eq_leib1]\n" h1;
		          | Lam(_,Lam(_,Ap(Forall(_),Lam(_,(Ap(Ap(Imp,Ap(Ap(Imp,(Ap(DB(0,_),DB(2,_)))),False)),Ap(Ap(Imp,(Ap(DB(0,_),DB(1,_)))),False)) ))) )) ->
                  Printf.fprintf c "%snote %s = eq_ind[THEN spec, of \"%% s t. ! p. ~ p s --> ~ p t\", THEN spec, of \"" sp h2;
				          trm_to_isar c prefix (Variables.make ());
                  Printf.fprintf c "\", THEN mp, OF %s, THEN spec, of \"(=)\", THEN mp, OF eq_leib2]\n" h1;
		          | Lam(_,Lam(_,Ap(Forall(_),Lam(_,(Ap(Ap(Imp,(Ap(Forall(_),Lam(_,(Ap(Ap(DB(1,_),DB(0,_)),DB(0,_)))))) ),(Ap(Ap(DB(0,_),DB(2,_)),DB(1,_)))) ) )) )) ->
                  Printf.fprintf c "%snote %s = eq_ind[THEN spec, of \"%% s t. ! r. (! x. r x x) --> r s t\", THEN spec, of \"" sp h2;
				          trm_to_isar c prefix (Variables.make ());
                  Printf.fprintf c "\", THEN mp, OF %s, THEN spec, of \"(=)\", THEN mp, OF eq_leib3]\n" h1;
		          | Lam(_,Lam(_, Ap(Forall(_),Lam(_,(Ap(Ap(Imp,(Ap(Ap(Imp,(Ap(Ap(DB(0,_),DB(2,_)),DB(1,_)))),False) )),(Ap(Ap(Imp,(Ap(Forall(_),Lam(_,(Ap(Ap(DB(1,_),DB(0,_)),DB(0,_))))) )),False) )) ) )) )) ->
                  Printf.fprintf c "%snote %s = eq_ind[THEN spec, of \"%% s t. ! r. r s t --> ~(! x. r x x)\", THEN spec, of \"" sp h2;
				          trm_to_isar c prefix (Variables.make ());
                  Printf.fprintf c "\", THEN mp, OF %s, THEN spec, of \"(=)\", THEN mp, OF eq_leib4]\n" h1;
		          | _ -> failwith("unknown rewrite step found in ref_coq" ^ (trm_str pt))
	        end;
	        ref_isabellehol1 c r1 ((s,h2)::hyp) const sp
    | Delta(h,s,x,r1) ->
	      let h1 = (lookup "29" (coqnorm h) hyp) in
	        Printf.fprintf c "%snote %s = %s[unfolded %s_def]\n" sp h1 h1 (Hashtbl.find coq_names x);
	        ref_isabellehol1 c r1 ((coqnorm s,h1)::hyp) const sp;
    | KnownResult(s,name,al,r1) ->
        let name = isabelle_itp_thm_str name in
        begin
          match al with
            | (_::_) ->
	              let h1 = get_hyp_name() in
	                Printf.fprintf c "%snote %s = %s" sp h1 name;
                  let length_al = List.length al
                  in
                    if length_al > 0 then
                      begin
	                      Printf.fprintf c "[where ";
                        let ty_names =
                          List.combine (countup 1 length_al []) al
                        in
	                        ignore(List.fold_right
                            (*"ty" is the name of the schematic type variable;
                              "a" is the object-level type it's being instantiate to*)
	                          (fun (ty, a) remaining -> (
                               (*NOTE the typename is suffixed with "_" otherwise isabelle will
                                 have trouble matching the intended schematic variable.. this is hackish*)
	                             Printf.fprintf c "'ty%s_ = " (string_of_int ty); (*FIXME const*)
	                             print_stp_isar c a true;
                               if remaining > 1 then Printf.fprintf c " and "; (*FIXME const*)
                               remaining - 1))
	                          ty_names length_al);
	                        Printf.fprintf c "]"
                      end;
	                  Printf.fprintf c "\n";
	                  ref_isabellehol1 c r1 ((coqnorm s,h1)::hyp) const sp
            | [] ->
	              ref_isabellehol1 c r1 ((coqnorm s,name)::hyp) const sp
        end;
    | NYI(h,s,r1) -> failwith("NYI step found in ref_coq" )
    | Timeout -> failwith("Timeout step found in ref_coq" )

let ref_isabellehol c r =
	(* get conjecture *)
	let con =
    match !conjecture with
        Some(con, _) -> con
      | None -> False in
	(* initialise hypotheses *)
	let hyp =
    List.fold_left
      (fun l (s, pt) ->
         (coqnorm pt, s) :: l ) [] !coqsig_hyp in
	let h1 = get_hyp_name() in
  Printf.fprintf c "\nproof (rule ccontr)\n";
  Printf.fprintf c "  assume %s : \"" h1;

  trm_to_isar c (coqnorm (neg con)) (Variables.make ());

  Printf.fprintf c "\"\n";
  Printf.fprintf c "  show False\n";
  Printf.fprintf c "    proof (rule ccontr)\n";

  ref_isabellehol1 c r ((neg (coqnorm con), h1) :: hyp) (!coqsig_const) "    ";
  Printf.fprintf c "    qed\n";
  Printf.fprintf c "  qed\n";
  Printf.fprintf c "end\n"
