open Args
open Atom
open Problem
open Term
open Search
open Refutation
open Flag
open Suche
open Translation
open Branch
open Refut

exception ProofTooBig of int

let max_proof_size = ref None

let set_max_proof_size s = max_proof_size := s


(** Statistic **)

type statistics =
  { size     : int
  ; depth    : int
  ; width    : int
  ; cuts     : int
  ; rewrites : int
  ; nyis     : int
  ; timeouts : int
  }

let comb_stat s1 s2 =
  { size     = s1.size + s2.size
  ; depth    = max s1.depth s2.depth
  ; width    = s1.width + s2.width
  ; cuts     = s1.cuts + s2.cuts
  ; rewrites = s1.rewrites + s2.rewrites
  ; nyis     = s1.nyis + s2.nyis
  ; timeouts = s1.timeouts + s2.timeouts
  }

let incr_stat s = { s with size = 1 + s.size; depth = 1 + s.depth }

let zero_stat =
  { size     = 0
  ; depth    = 0
  ; width    = 0
  ; cuts     = 0
  ; rewrites = 0
  ; nyis     = 0
  ; timeouts = 0
  }

let unit_stat =
  { zero_stat with
    size  = 1
  ; depth = 1
  ; width = 1
  }

(** [statistic] extract statistics from a refutation *)
let statistic f on_timeout =
  let rec go h = function
  | Conflict(_,_)
  | Fal(_)
  | NegRefl(_) -> unit_stat
  | Implication(_,_,_,r1,r2)
  | Disjunction(_,_,_,r1,r2)
  | NegConjunction(_,_,_,r1,r2)
  | Confront(_,_,_,_,_,_,r1,r2)
  | NegEqualProp(_,_,_,r1,r2)
  | EqualProp(_,_,_,r1,r2)
  | NegAequivalenz(_,_,_,r1,r2)
  | Aequivalenz(_,_,_,r1,r2)
  | ChoiceR(_,_,_,_,r1,r2) ->
      let s1 = go (h+1) r1
      and s2 = go (h+1) r2 in
      f h (incr_stat (comb_stat s1 s2))
  | NegImplication(_,_,_,r1)
  | Conjunction(_,_,_,r1)
  | NegDisjunction(_,_,_,r1)
  | All(_,_,r1,_,_,_)
  | NegAll(_,_,r1,_,_,_)
  | Exist(_,_,r1,_,_,_)
  | NegExist(_,_,r1,_,_,_)
  | Trans(_,_,_,r1)
  | NegEqualFunc(_,_,r1)
  | EqualFunc(_,_,r1)
  | DoubleNegation(_,_,r1)
  | Delta(_,_,_,r1)
  | KnownResult(_,_,_,r1) ->
      let s1 = go (h+1) r1  in
      f h (incr_stat s1)
  | Cut(l,r1,r2) ->
      if debug_litcount then Printf.printf "cut on %d\n" (get_literal l);
      let s1 = go (h+1) r1
      and s2 = go (h+1) r2 in
      let s = incr_stat (comb_stat s1 s2) in
      f h ({ s with cuts = 1 + s.cuts })
  | Rewrite(_,_,_,r1) ->
      let s1 = go (h+1) r1 in
      let s = incr_stat s1 in
      f h { s with rewrites = 1 + s.rewrites }
  | NYI(_,_,r1) ->
      let s1 = go (h+1) r1 in
      let s = incr_stat s1 in
      f h { s with nyis = 1 + s.nyis }
  | Mating(_,_,_,[])
  | Decomposition(_,_,[]) -> failwith "Refutation list is empty"
  | Mating(_,_,_,rs)
  | Decomposition(_,_,rs) ->
      let sl = List.map (go (h+1)) rs in
      let s = List.fold_left comb_stat zero_stat sl in
      f h (incr_stat s)
  | Timeout -> on_timeout h
  in go 0

let statistic_statcount r =
  (* [statcount] is an attempt to guess the size of the final refutation
     after timeout stopped the search *)
  let statcount = Hashtbl.create 100 in
  let update_statcount h s =
    let (zs, zw, n) = try Hashtbl.find statcount h with Not_found -> (0, 0, 0) in
    Hashtbl.replace statcount h (zs + s.size, zw + s.width, n + 1);
    s
  and handle_timeout h =
    let (zs, zw, n) = try Hashtbl.find statcount h with Not_found -> (1, 1, 1) in
    { zero_stat with size = zs/n; depth = 1; width = zw/n; timeouts = 1 }
  in statistic update_statcount handle_timeout r

let statistic r =
  let s = statistic_statcount r in
  (s.size, s.size, s.depth, s.width, s.width, s.cuts, s.rewrites, s.nyis, s.timeouts)



let print_latex_search (size,_,depth,width,_,cut,rewrite,notyetImpl,_) lat c =
  if (result_latex && width < 50 && depth < 30) then
  Printf.fprintf c "(*** \n %%beginLatex \n size %d depth %d width %d \n \n \\begin{tabular}{c} \n %s \\end{tabular} \n\n %%endLatex \n ***)\n" 
  size depth width lat
  else if (result_latex) then Printf.fprintf c "(*** \n %%beginLatex \n size %d depth %d width %d \n \n  %%endLatex \n ***)\n" size depth width


let print_latex_translation (size,_,depth,width,_,cut,rewrite,notyetImpl,_) lat c =
  if (result_latex &&  width < 50 && depth < 30) then
  Printf.fprintf c "(*** \n %%beginLatex \n size %d depth %d width %d cuts %d rewrite  %d NYI %d \n \n \\begin{tabular}{c} \n %s \\end{tabular} \n\n %%endLatex \n***)" 
  size depth width cut rewrite notyetImpl lat
  else if (result_latex) then Printf.fprintf c "(*** \n %%beginLatex \n \n \n Translation successful, probably \n \n size %d depth %d width %d cuts %d rewrite %d  NYI %d \n %%endLatex \n***)" size depth width cut rewrite notyetImpl 
  ; flush stdout


let print_refutation_result (refutation,_,searchTime,isTimeout) =
  let (size,_,depth,width,_,cut,rewrite,notyetImpl,timeouts) = statistic refutation in
  let hasTimedout = isTimeout || timeouts > 0 in
  if hasTimedout then (if !verbosity > 0 then Printf.printf  "Timeout!  time %d ms size %d depth %d width %d cuts %d rewrite %d NYI %d timeouts %d \n" searchTime size depth width cut rewrite notyetImpl timeouts)
  else
  begin
    if result_print_search then Printf.printf  "%s\n" (ref_str refutation);
    if (!verbosity > 3) then Printf.printf "Search time %d ms size %d depth %d width %d \n" searchTime size depth width;
  end;
  (refutation, hasTimedout)
  
let get_initbranch () =
  let initbranch1 = (List.rev !initial_branch) in
  List.map preprocess initbranch1 (*** Chad, Aug 2011: Preprocess to handle LEIBEQ_TO_PRIMEQ here as well as in suche.ml. Example: SYO348^5.p ***)

let get_initbranch_prenorm () =
  let go pn =
    let p = onlynegnorm pn in
    if debug_translation then Printf.printf "%s = %s \n" (trm_str pn) (trm_str p); p in
  List.map go (List.rev !initial_branch_prenorm)
  

let get_refutation initbranch r =
  (* get assumptions *)
  let b = List.map get_literal initbranch in
  (* call search *)
  let (_,con,_,_) as sr = search_refutation b r in
  if assert_condition && not (Dependency.is_empty (Dependency.diffunion [] [con] [b])) && !verbosity > 0 then
    Printf.printf "Error with refutation: Still open conditions\n";
  sr

let do_translate branch refutation =
      try
	if pftrm_eager_rewrite then
	  eager_translate branch refutation
	else
	  lazy_translate branch refutation
      with TranslateFailure -> safe_translate branch refutation (*** Chad Aug 2011 : Fall back on safe_translate ***)

let get_translation branch refutation =
  if debug_translation then (Printf.printf "About to translate. Branch assoc with prenorm:\n"; List.iter (fun (m,p) -> Printf.printf "trm: %s\npre: %s\n" (trm_str m) (trm_str p)) branch);
(* The flag pftrm_eager_rewrite decides wether rewrites are eagerly or lazily handled *) 
  let prenorm_refutation = do_translate branch refutation in

  if debug_translation then (Printf.printf "Translated prenorm_refutation\n%s\n" (ref_str prenorm_refutation));
  prenorm_refutation

let print_prenorm_ref_result transTime stats prenorm_refutation =
  let (size,_,depth,width,_,cut,rewrite,notyetImpl,_) as stats = statistic prenorm_refutation  in
  if result_print_translation then Printf.printf  "%s\n" (ref_str prenorm_refutation);
  if (!verbosity > 3) then Printf.printf  "Translation NYI %d time %d ms size %d depth %d width %d cuts %d rewrite %d  \n" notyetImpl transTime size depth width cut rewrite

let assert_proof_size size =
  match !max_proof_size with
  | Some max when size > max -> raise (ProofTooBig size)
  | _ -> ()

let print_coq_proof printer c r =
  if (!verbosity > vold) then print_refut r;

	(*** Search ***)

  let initbranch = get_initbranch () in
  let (refutation, hasTimedout) = print_refutation_result (get_refutation initbranch r) in

  if hasTimedout then raise Timeout.Timeout;
	(*** Translation  ***)

(* get prenormalized assumptions *)
  let initbranch_prenorm = get_initbranch_prenorm () in

  let beforeTrans= Sys.time() in
(* Branch is an association list from normalized terms to their prenormalized counterparts *)
  let branch = List.combine initbranch initbranch_prenorm in
  let prenorm_refutation = get_translation branch refutation in
  let transTime= int_of_float ((Sys.time() -. beforeTrans) *. 1000.0) in

  let (size, _,_,_,_,_,_,_,_) as stats = statistic prenorm_refutation in
  print_prenorm_ref_result transTime stats prenorm_refutation;

  (*** Output Coq ***)

  let beforeCoq = Sys.time() in
  assert_proof_size size;
  printer c prenorm_refutation;
  let coqTime = int_of_float ((Sys.time() -. beforeCoq) *. 1000.0) in
  if !verbosity > 3 then Printf.printf "Coq output done: time %d\n" coqTime;

  print_latex_search stats (Latex.ref_to_lat initbranch refutation) c;
  print_latex_translation stats (Latex.ref_to_lat initbranch_prenorm prenorm_refutation) c


let print_coq_proofscript c r =
  if (!verbosity > vold) then Printf.printf  "starting print_coq_proofscript.\n";
  print_coq_proof Coq.ref_coq c r

let print_isar_proofscript c r =
  if (!verbosity > vold) then Printf.printf  "starting print_isar_proofscript.\n";
  print_coq_proof Isabellehol.ref_isabellehol c r

let print_coq_sproofterm c r =
  if (!verbosity > vold) then Printf.printf "starting print_coq_proofspfterm.\n";
  print_coq_proof Coqspfterm.ref_coq_spfterm c r

let print_dedukti c r =
  print_coq_proof Dedukti.ref_dedukti c r

let refut_trms r =
  let (refutation,_,_,_) = get_refutation (get_initbranch ()) r in
  ref_trms refutation
