(* File: syntax.ml *)
(* Author: Chad E Brown *)
(* Created: March 2008/September 2008 *)

exception ParsingError of int * int * int * int * string
exception GenericSyntaxError of string
exception DeclareInd
exception Polymorphism

type pretrm =
    PName of string
  | PType | PPropTp | PIndTp | PIntTp | PRealTp | PTrue | PFalse | PIff | PNIff | PImplies | PRevImplies | POr | PNOr | PAnd | PNAnd | PEq | PNEq | PNeg | PForall | PExists
  | PInt of int
  | PReal of float
  | PAp of pretrm * pretrm
  | PULam of string list * pretrm
  | PLam of ((string * pretrm) list) * pretrm
  | PAll of ((string * pretrm) list) * pretrm
  | PEx of ((string * pretrm) list) * pretrm
  | PExU of string * pretrm * pretrm
  | PChoice of (string * pretrm) * pretrm
  | PAr of pretrm * pretrm
  | POf of pretrm * pretrm
  | PDef of pretrm * pretrm
  | PLet of string * pretrm * pretrm
  | PTLet of string * pretrm * pretrm * pretrm

type prectx = (string option * pretrm option * pretrm option) list

let rec prectx_lam c m =
  match c with
  | (Some x,Some a,None)::c' -> prectx_lam c' (PLam([(x,a)],m))
  | (Some x,_,Some n)::c' -> prectx_lam c' (PLet(x,n,m))
  | (None,Some p,None)::c' -> prectx_lam c' m
  | _::_ -> raise (GenericSyntaxError "Impossible prectx case")
  | [] -> m

let rec prectx_all c m =
  match c with
  | (Some x,Some a,None)::c' -> prectx_all c' (PAll([(x,a)],m))
  | (Some x,_,Some n)::c' -> prectx_all c' (PLet(x,n,m))
  | (None,Some p,None)::c' -> prectx_all c' (PAp(PAp(PImplies,p),m))
  | _::_ -> raise (GenericSyntaxError "Impossible prectx case")
  | [] -> m

let rec prectx_ar c m =
  match c with
  | (Some x,Some a,None)::c' -> prectx_ar c' (PAr(a,m))
  | (Some x,_,Some n)::c' -> prectx_ar c' m
  | (None,Some p,None)::c' -> prectx_ar c' m
  | _::_ -> raise (GenericSyntaxError "Impossible prectx case")
  | [] -> m

let rec pretrm_str m =
  match m with
  | PName x        -> x
  | PType          -> "$type"
  | PPropTp        -> "$o"
  | PIndTp         -> "$i"
  | PIntTp         -> "$int"
  | PRealTp        -> "$real"
  | PTrue          -> "$true"
  | PFalse         -> "$false"
  | PIff           -> "$iff"
  | PNIff          -> "$niff"
  | PImplies       -> "$implies"
  | PRevImplies    -> "$revimplies"
  | POr            -> "$or"
  | PNOr           -> "$Nor"
  | PAnd           -> "$and"
  | PNAnd          -> "$Nand"
  | PEq            -> "$eq"
  | PNEq           -> "$neq"
  | PNeg           -> "~"
  | PForall        -> "!!"
  | PExists        -> "??"
  | PInt i         -> string_of_int i
  | PReal z        -> string_of_float z
  | PAp(n,p)       -> "(" ^ pretrm_str n ^ " @ " ^ pretrm_str p ^ ")"
  | PULam(xl,m)    -> "(^ [" ^ String.concat "," xl ^ "] " ^ pretrm_str m ^ ")"
  | PLam (xl,m)    -> "(^ [" ^ typings xl ^ "] " ^ pretrm_str m ^ ")"
  | PAll(xl,m)     -> "(! [" ^ typings xl ^ "] " ^ pretrm_str m ^ ")"
  | PEx(xl,m)      -> "(? [" ^ typings xl ^ "] " ^ pretrm_str m ^ ")"
  | PExU(x,a,m)    -> "(?! [" ^ typing (x, a) ^ "] " ^ pretrm_str m ^ ")"
  | PChoice(xa,m)  -> "(@+ [" ^ typing xa ^ "] " ^ pretrm_str m ^ ")"
  | PAr(n,p)       -> "(" ^ pretrm_str n ^ " > "  ^ pretrm_str p ^ ")"
  | POf(n,p)       -> "(" ^ pretrm_str n ^ " : "  ^ pretrm_str p ^ ")"
  | PDef(n,p)      -> "(" ^ pretrm_str n ^ " := " ^ pretrm_str p ^ ")"
  | PLet(x,n,p)    -> "(" ^ x ^ " := " ^ pretrm_str n ^ " in " ^ pretrm_str p ^ ")"
  | PTLet(x,a,n,p) -> "(" ^ typing (x, a) ^ " := " ^ pretrm_str n ^ " in " ^ pretrm_str p ^ ")"
and typing (x, a) = x ^ " : " ^ pretrm_str a
and typings xl = String.concat "," (List.map typing xl)

type input =
    DeclareTHF of string * string * pretrm
  | Include of string
