open Args
open Error
open Flags
open Ontology
open Preterm
open Term
open TermP


let coq = ref false
let translucent_defns = ref false

type probitem =
  | ProbDef of string * stp * trm * (string * string) list * float
  | ProbAx of string * string * trm * (string * string) list * float
  | ProbConj of string * trm * (string * string) list * float
let probitem_weight = function
    ProbDef(_,_,_,_,w) -> w
  | ProbAx (_,_,_,_,w) -> w
  | ProbConj (_,_,_,w) -> w
let probsig : probitem list ref = ref []

let conjecturename : string ref = ref "claim"
let conjecture : (trm * trm) option ref = ref None

let coqsig_base : string list ref = ref []
let coqsig_const : (string * stp) list ref = ref []
let coqsig_def : (string * trm) list ref = ref []
let coqsig_hyp : (string * trm) list ref = ref []

let name_base : (string,unit) Hashtbl.t = Hashtbl.create 10
let name_base_list : string list ref = ref []
let name_tp : (string,stp) Hashtbl.t = Hashtbl.create 100
let name_trm : (string,(trm * stp) * bool ref) Hashtbl.t = Hashtbl.create 100
let name_trm_list : (string * trm * stp) list ref = ref []
let name_def_empty : (string,trm) Hashtbl.t = Hashtbl.create 1
let name_def : (string,trm) Hashtbl.t = Hashtbl.create 100
let name_def_all : (string,trm) Hashtbl.t = Hashtbl.create 100
let name_def_prenorm : (string,trm) Hashtbl.t = Hashtbl.create 100
let name_hyp : (string,trm) Hashtbl.t = Hashtbl.create 100
let name_hyp_inv : (trm,string * trm) Hashtbl.t = Hashtbl.create 100 (*** associate a normalized/logic normalized term with the name of the assumption it came from and its original prenormalized form ***)

let initial_branch : trm list ref = ref []
let initial_branch_prenorm : trm list ref = ref []

let part_of_conjecture : (trm,unit) Hashtbl.t = Hashtbl.create 10

(** applies neg-, beta- and delta-normalization**)
let coqnorm m =
  let m = betanorm name_def_prenorm m in
  let (n,_) = negnorm1 m in n
(** partially normalizes, without expanding certain defns (e.g., is_of) **)
let partialnormalize pt = norm name_def (logicnorm pt)
(** applies full satallax normalization**)
let normalize pt = norm name_def_all (logicnorm pt)
let belnorm pt = norm name_def_empty (logicnorm pt)


let coqify_name_1 x =
  let r = ref "" in
  if (x = "$i") then r := "i" else
  begin
    let fst = ref true in
    String.iter
      (fun c ->
	let co = Char.code c in
	if (((co >= 65) && (co <= 90)) || ((co >= 97) && (co <= 122))) then
	  begin
	    fst := false;
	    r := (!r) ^ (Char.escaped c)
	  end
	else
	  begin
	    if ((!fst) && (co <> 95)) then r := "_";
	    fst := false;
	    if ((co >= 48) && (co <= 57)) then
	      r := (!r) ^ (Char.escaped c)
	    else if (co = 39) then
	      r := (!r) ^ "'"
	    else if (co = 95) then
	      r := (!r) ^ "_";
	    (*** omit any other characters, relying on coqify_name_0 to ensure it's fresh ***)
	  end
      )
      x;
    if ((!r) = "_") then r := "_X";
  end;
  r

let coqify_name_0 x h =
  let r = coqify_name_1 x in
  while (Hashtbl.mem h (!r)) do
    r := (!r) ^ "'"; (*** add primes until it's fresh ***)
  done;
  (!r)

let coqify_name x h1 h2 =
  let y = coqify_name_0 x h2 in
  Hashtbl.add h1 x y;
  Hashtbl.add h2 y ();
  y



let declare_name_type x tp =
  let trm = Name(x,tp) in
  Hashtbl.add name_tp x tp;
  Hashtbl.add name_trm x ((trm, tp), ref false);
  name_trm_list := (x, trm, tp)::!name_trm_list

(*** March 31, 2011 - Chad - THF policy regarding definitions. (See comments before declare_definition_real below. ***)
exception TreatAsAssumption

let next_fresh_name : int ref = ref 0

let rec get_fresh_name a =
  let x = make_fresh_name !next_fresh_name in
  incr next_fresh_name;
  if !coq then ignore (coqify_name x coq_names coq_used_names);
  if ((Hashtbl.mem name_base x) || (Hashtbl.mem name_tp x)) then
    get_fresh_name a
  else
    begin
      declare_name_type x a;
      (x, Name(x, a))
    end

let default_elts : (string,trm) Hashtbl.t = Hashtbl.create 10

let set_default_elt aname x = Hashtbl.add default_elts aname x

let set_default_elt_if_needed a x =
  match a with
  | Base(aname) ->
      if (not (Hashtbl.mem default_elts aname)) then set_default_elt aname x
  | _ -> ()

let default_elt aname =
  try
    Hashtbl.find default_elts aname
  with
  | Not_found ->
      let a = Base aname in
      let (_,x) = get_fresh_name a in
      begin
	set_default_elt aname x;
	x
      end

let default_elt_p aname =
  Hashtbl.mem default_elts aname



let declare_base_type a =
  if (!coq) then
    begin 
      let y = coqify_name a coq_names coq_used_names in
      coqsig_base := (y::!coqsig_base)
    end;
  Hashtbl.add name_base a ();
  name_base_list := (a::!name_base_list);
  if (get_bool_flag "CHOICE_AS_DEFAULT") then
    set_default_elt a (norm name_def (ap(Choice (Base a),Lam(Base a,False))))


(* Declare the base type i the first time it's seen.
 * Use a different name if an 'i' has already been used.
 *)
let with_declare_ind f =
  try f () with DeclareInd -> (declare_base_type "$i"; f ())

let st_to_stp m = with_declare_ind (fun () -> to_stp m)

let st_to_trm m = with_declare_ind (fun () -> to_trm name_trm [] m None)

let st_to_trm_given_stp m tp = with_declare_ind
  (fun () -> let (n,_) = to_trm name_trm [] m (Some tp) in n)



let declare_typed_constant (name:string) (role:string) (m:pretrm) (al:(string * string) list) =
  if (!verbosity > 4) then (Printf.printf "declare_typed_constant %s %s\n%s\n" name role (pretrm_str m); flush stdout);
  if (!verbosity > 20) then (Printf.printf "annotations:\n"; List.iter (fun (a,b) -> Printf.printf "%s: %s\n" a b) al; flush stdout);
  let check_redeclaration x tp =
    (* fail if x has been defined with a different type before *)
    try
      let tp' = Hashtbl.find name_tp x in
      if tp <> tp' then raise (Redeclaration x)
    with Not_found ->
      if Hashtbl.mem name_base x then raise (Redeclaration x) in
  match m with
    POf(PName(x),PType)
  | POf(PName(x),PName "$type") ->
      if Hashtbl.mem name_tp x then raise (Redeclaration x);
      declare_base_type x
  | POf(PName(x),m) ->
      let tp = st_to_stp m in
      add_ontology_type x tp al;
      check_redeclaration x tp;
      if (!coq) then
      begin
        let y = coqify_name x coq_names coq_used_names in
        coqsig_const := (y,tp)::!coqsig_const
      end;
      declare_name_type x tp
  | _ -> raise (GenericError ("Incorrect format for type declaration " ^ name))

let rec init_probitem m =
  match m with
  | ProbDef(x,a,m,al,w) ->
      begin
	try
	  if (get_bool_flag "ALL_DEFS_AS_EQNS") then
	    raise TreatAsAssumption
	  else
	  begin
	    Hashtbl.add name_def_prenorm x m;
	    let m0 = logicnorm m in
	    let m2 = norm name_def m0 in
	    if (!verbosity > 20) then Printf.printf "name_def %s %s\n" x (trm_str m2);
	    Hashtbl.add name_def_all x (norm name_def_all m2);
	    if not (translucent_defn_p m) then
	      begin
		(* TODO: really set translucent definitions to true? *)
		translucent_defns := true;
		Hashtbl.add name_def x m2
	      end
	  end
	with TreatAsAssumption ->
	  init_probitem (ProbAx(x,"hypothesis",Ap(Ap(Eq(a),Name(x,a)),m),al,w))
      end
  | ProbAx(name,role,tm,al,w) ->
      begin
	let m0 = logicnorm tm in
	let tmn = norm name_def m0 in
	Hashtbl.add name_hyp name tmn;
	Hashtbl.add name_hyp_inv tmn (name,tm);
	if (!verbosity > 20) then Printf.printf "name_hyp %s %s\n" name (trm_str tmn);
	initial_branch_prenorm := tm::(!initial_branch_prenorm);
	initial_branch := tmn::(!initial_branch)
      end
  | ProbConj(name,tm,al,w) ->
      begin
	conjecturename := name;
	if (name != "claim") then
	  ignore (coqify_name name coq_hyp_names coq_used_names);
	let ntm = Ap(Neg,tm) in
	let ntmn = norm name_def (logicnorm ntm) in
	initial_branch_prenorm := (ntm::(!initial_branch_prenorm));
	if (!verbosity > 20) then Printf.printf "name_conj negated %s %s\n" name (trm_str ntmn);
	initial_branch := (ntmn::(!initial_branch));
	conjecture := Some (tm,ntmn);
	Hashtbl.add part_of_conjecture ntmn ()
      end

let defpreweight m al = 2.0

let axpreweight m al =
  if ontology_prop_p m then 1.0 else 3.0

let conjpreweight m al = 0.0

(***
 THF official policy requires that all names are given a type before use,
 that definitions can be circular (in which case they must be treated as equations),
 and that definitions may appear *after* it's already been used in an assumption.
 In order to comply with this policy *and* do something reasonable in the 'usual' case,
 I do the following:

 1. I keep a hashtable (occurred_names) of typed names that have been used in an assumption.
 2. When I encounter a definition, if it's of the form (x = t), I first parse t.  If x has been used,
    then I include the equation as an axiom.  Otherwise, x is treated as a definition and expanded away.

 NOTE: If I wanted to be strict, I would exit if the definition is given without the name being given a type,
 but I will not require the type to be given.  I will print a warning if the verbosity is > 0.
 ***)
let declare_definition_real x m al =
  if (Hashtbl.mem name_base x) then raise (Redeclaration x);
  if (Hashtbl.mem name_def x) then raise TreatAsAssumption; (*** treat it as an assumption per THF policy. - Mar 31, 2011 - Chad ***)
(*** raise (Redeclaration x); ***)
  try
    let tp = Hashtbl.find name_tp x in
    let tm = belnorm (st_to_trm_given_stp m tp) in
    try 
      let (_,r) = Hashtbl.find name_trm x in
      if (!r) then
	raise TreatAsAssumption (*** It's already been used, treat 'definition' as an assumption. ***)
      else
	raise Not_found
    with
    | Not_found ->
	begin
	  if (!coq) then (*** Coq proofs now only work for non-slaves ***)
	    begin (*** The name was coqified when the type was declared. ***)
	      try
		let y = Hashtbl.find coq_names x in
		coqsig_def := (y,tm)::!coqsig_def
	      with
	      | Not_found -> raise (GenericError("Could not find Coq version of name " ^ x))
	    end;
	  add_ontology_term x tm;
	  let w = defpreweight tm al in
	  probsig := ProbDef(x,tp,tm,al,w)::!probsig
	end
  with
  | Not_found ->
      begin (*** Giving a definition without giving it's type isn't allowed in THF anymore.  I'm still allowing it.  ***)
	if ((!verbosity > 0) (*&& (not (!coqglobalfile)) && (not (!coqlocalfile))*)) then Printf.printf "WARNING: %s defined without giving type first.\n" x;
	let (tm,tp) = st_to_trm m in
	let tm = belnorm tm in
	if (!coq) then
	  begin
	    let y = coqify_name x coq_names coq_used_names in
	    coqsig_const := (y,tp)::!coqsig_const;
	    coqsig_def := (y,tm)::!coqsig_def
	  end;
	declare_name_type x tp;
	let w = defpreweight tm al in
	probsig := ProbDef(x,tp,tm,al,w)::!probsig
      end

let declare_thf_logic_formula (name:string) (role:string) (m:pretrm) (al:(string * string) list) =
  begin
    if !verbosity > 4 then (Printf.printf "declare_thf_logic_formula %s %s\n" name role; flush stdout);
    if (!verbosity > 20) then (Printf.printf "annotations:\n"; List.iter (fun (a,b) -> Printf.printf "%s: %s\n" a b) al; flush stdout);
    if ((role = "axiom") || (role = "hypothesis") || (role = "assumption") || (role = "lemma") || (role = "theorem") || (role = "corollary")) then
      begin
	if (Hashtbl.mem name_hyp name) then raise (Redeclaration name);
	let tm = st_to_trm_given_stp m Prop in
	let tm = belnorm tm in
	if (!coq) then (** Giving Coq proofs will now only work for non-slaves since the pretrm is used **)
	  begin
	    let y = coqify_name name coq_hyp_names coq_used_names in
	    coqsig_hyp := ((y,tm)::!coqsig_hyp)
	  end;
	let w = axpreweight tm al in
	probsig := ProbAx(name,role,tm,al,w)::!probsig
      end
    else if (role = "conjecture") then
      begin
	match (!conjecture) with
	| Some _ -> raise (GenericError "Problem file has more than one conjecture.")
	| None ->
	    let tm = st_to_trm_given_stp m Prop in
	    let tm = belnorm tm in
	    let w = conjpreweight tm al in
	    probsig := ProbConj(name,tm,al,w)::!probsig
      end
    else
      raise (GenericError ("Unknown role " ^ role))
  end

let declare_definition (name:string) (role:string) (m:pretrm) (al:(string * string) list) =
  try
    begin
      if (get_bool_flag "ALL_DEFS_AS_EQNS") then raise TreatAsAssumption;
      if !verbosity > 4 then (Printf.printf "declare_definition %s %s\n" name role; flush stdout);
      if (!verbosity > 20) then (Printf.printf "annotations:\n"; List.iter (fun (a,b) -> Printf.printf "%s: %s\n" a b) al; flush stdout);
      match m with
	PDef(PName(x),m) -> (*** No longer THF syntax. ***)
	  declare_definition_real x m al
      | PAp(PAp(PEq,PName(x)),m) ->
	  declare_definition_real x m al
      | _ -> (*** Treat as an assumption, no matter how it looks. Odd, but OK. This may be too liberal; we haven't decided yet. ***)
	  raise TreatAsAssumption
(*** raise (GenericError ("Incorrect format for definition " ^ name)) ***)
    end
  with
  | TreatAsAssumption ->
      declare_thf_logic_formula name "hypothesis" m al
