(* File: satallaxmain.ml *)
(* Author: Chad E Brown *)
(* Created: September 2010 - moved most of this from satallax.ml to satallaxmain.ml in September 2011 *)

open Args
open Flags
open Term
open Search
open Error
open Problem
open Sine
open Ontology


module Read = struct

let with_include = Tptp_config.with_include

let read_thf_file filename =
  let c = if filename = "" then stdin else open_in filename in
  let lex = Lexing.from_channel c in
  ignore (Tptp_parser.tptp_file Tptp_lexer.token lex);
  if (!verbosity > 4) then Printf.printf "Finished reading thf file %s\n" filename

let rec find_read_thf_file_r odir dir f =
  let ff = (dir ^ "/" ^ f) in
  if (Sys.file_exists ff) then
    with_include (find_read_thf_file odir) read_thf_file ff
  else if (String.length dir > 1) then
    find_read_thf_file_r odir (Filename.dirname dir) f
  else
    raise (FileNotFound f)
and find_read_thf_file dir f =
  let ff = (dir ^ "/" ^ f) in
  if (Sys.file_exists ff) then
    with_include (find_read_thf_file dir) read_thf_file ff
  else
    begin
      try
	let tptpdir = Sys.getenv "TPTP" in
	let tff = (tptpdir ^ "/" ^ f) in
	if (Sys.file_exists tff) then
          with_include (find_read_thf_file dir) read_thf_file tff
	else
	  find_read_thf_file_r dir dir f
      with
      | Not_found -> find_read_thf_file_r dir dir f
    end;;

let read_problem file =
  (* TODO: move coq stuff somewhere else *)
  if !coq then TermP.termP_init();
  Tptp_config.st_find_read_thf_fun := find_read_thf_file;
  with_include (find_read_thf_file (Filename.dirname file)) read_thf_file file;
  probsig := List.rev !probsig

end

module Proof = struct

let max_proof_size = function
  | CoqScript
  | CoqSPfTerm -> Some 10000
  | IsarScript -> Some 800
  | _ -> None

let requires_refut = function
  | TSTP
  | CoqScript
  | CoqSPfTerm
  | Dedukti
  | IsarScript -> true
  | _ -> false

let prepare_proofterm pt =
  Proofterm.set_max_proof_size (max_proof_size pt);
  if requires_refut pt then coq := true

let code_status = function
  | (true , Some _) -> 10, "CounterSatisfiable"
  | (true , None  ) -> 15, "Satisfiable"
  | (false, Some _) -> 20, "Theorem"
  | (false, None  ) -> 25, "Unsatisfiable"

let n_inferences () = Queue.length Searchoption.searchoptions_retrieved
let inferences_str () = "% Inferences: " ^ string_of_int (n_inferences ())

let show_status s =
  [ "% SZS status " ^ s
  ; "% Mode: " ^ (String.concat " " !mode)
  ; inferences_str ()
  ]

let print_status s =
  if !verbosity > 0 then List.iter print_endline (show_status s)

let print_proofmsg c l =
  let enbracket s =
    if !mkproofterm = Some IsarScript then "(*" ^ s ^ "*)" else s in
  if c = stdout then List.iter print_endline (List.map enbracket l)

let print_start c l =
  let (_, status) = code_status (false, !conjecture) in
  print_proofmsg c (show_status status @ l)

let print_end c l = print_proofmsg c l; if c != stdout then close_out c else flush c

let try_proofout f =
  try f ()
  with Proofterm.ProofTooBig coqproofsize ->
    if (!verbosity > 0) then Printf.printf "%% SZS status Success\nProof Too Big: %d steps\n" coqproofsize;
    exit 26

let print_proofterm_full file r pt =
  let c = match file with
    None -> stdout
  | Some file -> open_out file in
  match pt with
  | TSTP ->
      print_start c ["% SZS output start Proof"];
      try_proofout (fun () -> Tstp.refut_tstp c r);
      print_end c ["% SZS output end Proof"]
  | CoqScript ->
      Coq.coq_sig c;
      print_start c ["% SZS output start Proof"; "% Coq Proof Script"];
      try_proofout (fun () -> Proofterm.print_coq_proofscript c r);
      print_end c ["% SZS output end Proof"]
  | CoqSPfTerm ->
      Coqspfterm.coq_sig c;
      print_start c ["% SZS output start Proof"; "% Coq Proof Script"];
      try_proofout (fun () -> Proofterm.print_coq_sproofterm c r);
      print_end c ["% SZS output end Proof"]
  | IsarScript ->
      Isabellehol.isabellehol_sig c file;
      print_start c ["% SZS output start Proof"; "% Isar Proof Script"];
      try_proofout (fun () -> Proofterm.print_isar_proofscript c r);
      print_end c ["% SZS output end Proof"]
  | Dedukti ->
      Dedukti.dk_sig c;
      print_start c ["% SZS output start Proof"; "% Dedukti Proof Script"];
      try_proofout (fun () -> Proofterm.print_dedukti c r);
      print_end c ["% SZS output end Proof"]
  | HOCore ->
      print_start c ["% Higher-Order Unsat Core BEGIN"];
      Hocore.print_hocore c r;
      print_end c ["% Higher-Order Unsat Core END"]
  | PfInfo ->
      print_start c ["% Pf Info BEGIN"];
      Pfinfo.print_pfinfo c r;
      print_end c ["% Pf Info END"]
  | PfUseful ->
      print_start c ["% Pf Useful BEGIN"];
      Pfuseful.print_pfuseful c r !pfusefulout;
      print_end c ["% Pf Useful END"]
  | PfFormdeps ->
      print_start c ["% Pf Formdeps BEGIN"];
      Pfformdeps.print_pfformdeps c r !pfformdepsout;
      print_end c ["% Pf Formdeps END"]
  | _ -> ()

let handle_result = function
  | Some(r) -> (* unsatisfiable *)
      let (code, status) = code_status (false, !conjecture) in
      begin match (r, !mkproofterm) with
          (Some r, Some pt) ->
            print_proofterm_full !proofoutfile r pt; exit code
        | (_, _) -> print_status status; exit code
      end
  | None -> (* satisfiable *)
      let (code, status) = code_status (true, !conjecture) in
      print_status status; exit code

end

module Schedule = struct

let preweightsum ps = List.fold_left (fun v z -> v +. probitem_weight z) 0.0 ps

let problem_is_small ps = preweightsum ps < 128.0

let auto () =
  if (!nontheorem) then "schedule_nontheorem"
  else if problem_is_small !probsig then "schedule_3_1"
  else "schedule_3_1_sine"

let get = function
  [] -> Mode.load_schedule (auto ())
| st -> List.concat (List.rev_map Mode.load_schedule st)

end

module Slave = struct

let timeout_args = function
  | Some(modedur, Some(userdur)) ->
      if userdur > modedur then
        [  "-t"; string_of_float modedur
        ; "-ht"; string_of_float userdur]
      else
        ["-t"; string_of_float userdur]
  | Some(modedur, None) -> ["-t"; string_of_float modedur]
  | None -> []

let slave_cmd dur m arg =
  let mode = ["-m"; m] in
  let timo = timeout_args dur in
  String.concat " " ((List.rev !slaveargs) @ arg @ mode @ timo @ ["-preparsed"])

let write_preparsed c =
  output_value c !name_base_list;
  output_value c (List.map (fun (x,_,tp) -> (x,tp)) !name_trm_list);
  output_value c !is_of_names;
  output_value c !all_of_names;
  output_value c !probsig

let read_preparsed c =
  let basetps : string list = input_value c in
  List.iter declare_base_type (List.rev basetps);
  let nametrms : (string * stp) list = input_value c in
  List.iter (fun (x,tp) -> declare_name_type x tp) (List.rev nametrms);
  is_of_names := input_value c;
  List.iter (fun x -> Hashtbl.add is_of_name x ()) !is_of_names;
  all_of_names := input_value c;
  List.iter (fun x -> Hashtbl.add all_of_name x ()) !all_of_names;
  probsig := input_value c

(* print all output from channel to stdout *)
let to_stdout c =
  while true do
    let l = input_line c in
    Printf.printf "%s\n" l
  done

(** If the slave got a final result, then use it. *)
let handle_slave_return pstatus =
  match pstatus with
  | 10
  | 15
  | 20
  | 25 -> exit pstatus
  | 5 (* Timeout *)
  | 6 (* IncompleteSatisfiable *)
  | _ -> ()

(** start a slave *)
let run dur m =
  let final = match dur with Some _ -> false | None -> true in
  let cmd = slave_cmd dur m (if final then [] else ["-slave"]) in
  if (!verbosity > 4) then print_endline ("Starting slave: " ^ cmd);
  flush stdout; (*** 2015, to prevent race conditions with output of main process vs. slave process ***)
  let (myout,myin,myerr) = Unix.open_process_full cmd [| |] in
  write_preparsed myin;
  close_out myin;
  try to_stdout myout
  with End_of_file ->
    match Unix.close_process_full (myout,myin,myerr) with
    | (Unix.WEXITED pstatus) ->
	if (!verbosity > 4) then
	  Printf.printf "slave returned with status %d\n" pstatus;
	handle_slave_return pstatus; if final then exit pstatus
    | _ ->
	if (!verbosity > 0) then
	  begin
            print_endline "slave returned with unknown status";
            if final then exit 3
	  end

end

module Premsel = struct

let filter_probitems r ps =
  let ps' = List.filter (fun z -> probitem_weight z <= r) ps in
  if !verbosity > 4 then
    begin
      Printf.printf "Including those ranked <= %f\n" r;
      List.iter
        (fun z ->
          match z with
          | ProbDef(x,_,_,_,_) -> Printf.printf "Including def %s\n" x
          | ProbAx(x,_,_,_,_) -> Printf.printf "Including ax %s\n" x
          | ProbConj(x,_,_,_) -> Printf.printf "Including conj %s\n" x)
        ps'
    end;
  ps'

let sine ps =
  filter_probitems (get_float_flag "SINE_RANK_LIMIT") (sinelike_premsel ps)


end

module Testing = struct

let print_branch =
  List.iteri (fun i m -> Printf.printf "%d %s\n" (i+1) (trm_str m))

(* select_list l [i1; ...; in] = [nth l in; ...; nth l i1] *)
let select_list l = List.rev_map (List.nth l)

let select_axioms () =
  (*** I should select the prenorm branch too, but that's only needed for proof reconstruction. ***)
  initial_branch := select_list !initial_branch !select_axioms_list;
  if (!verbosity > 4) then
    begin
      print_endline "Initial branch after selection:";
      print_branch !initial_branch
    end

let print_num_axioms () =
  let num_axioms = List.length !initial_branch in
  Printf.printf "(NUMAXIOMS \"%s\" %d)\n" !problemfile num_axioms;
  exit 123

end


(* TODO: remove code duplication with adapt_timeout from search.ml *)
let set_timeouts s =
  if (s > 0.0) then begin
    if (!nontheorem && get_bool_flag "SPLIT_GLOBAL_DISJUNCTIONS" && s >= 0.2)
    then (Timeout.set_timeout (s *. 0.5); OptionE.map_ref (( *.) 0.5) timeout)
    else (Timeout.set_timeout s; timeout := Some 0.0)
  end


(** total duration of schedule *)
let schedule_duration = List.fold_left (fun s1 (x,s2) -> (s1 +. s2)) 0.0

(** there are several times involved:
    * [usertime]: absolute time when timeout given by user is reached
    *  [userdur]: duration until timeout given by user is reached
    *   [schdur]: duration until final time of schedule is reached
    *  [modedur]: duration for which mode should be running *)
let rec run_schedule usertime schdur = function
  | [] -> raise Timeout.Timeout (* If nothing on schedule, then Timeout anyway. *)
  | [(m,_)] when usertime = None ->
      (* Last mode on timeoutless schedule, call it without a timeout and
         without telling it it's a slave. *)
      Slave.run None m
  | ((m, modedur)::sch) ->
      let dur =
        begin match usertime with
        | None -> (modedur, None)
        | Some usertime ->
            let userdur = usertime -. Unix.gettimeofday () in
            if userdur < modedur then
              raise Timeout.Timeout
            else
              (* if the user specified a timeout that is shorter than
                 the schedule duration, run each mode for shorter time *)
              let modedur = modedur *. (max 1. (userdur /. schdur)) in
              (modedur, Some userdur)
        end in
      Slave.run (Some dur) m;
      run_schedule usertime (schdur -. modedur) sch

let run_modes modes =
  try
    List.iter (fun mo -> Mode.load_mode_overrides (Mode.parse_mode_overrides mo)) modes;
    List.iter (fun (k, v) -> set_flag k v) !flag_overrides;
    if (!verbosity > 8) then print_flags ();
    OptionE.mapm Proof.prepare_proofterm !mkproofterm;
    set_timeouts (match !timeout with Some t -> t | None -> 0.0);
    if !preparsed then Slave.read_preparsed stdin else Read.read_problem !problemfile;
    if get_bool_flag "USE_SINE" then (probsig := Premsel.sine !probsig; completep := false);
    List.iter init_probitem !probsig;
    if (!printnumaxiomsflag) then Testing.print_num_axioms ();
    if (!selectaxiomsflag) then Testing.select_axioms ();
    Eprover.setup_eprover ();
    search ()
  with
  | Refut.Unsatisfiable(r) ->
      (*** Some subgoals may have timed out and the last one reported Unsatisfiable ***)
      if (!nontheorem) then raise Timeout.Timeout;
      (*** change the timeout to be the hard timeout, so that instead of continuing with the strategy schedule the rest of the time is spent trying to output this proof ***)
      begin
	match !hardtimeout with
	| Some hs ->
	    let rm = Timeout.remaining_time () in
	    Timeout.set_timeout (hs +. rm)
	| _ -> ()
      end;
      Proof.handle_result (Some r)
  | Refut.Satisfiable ->
      if not !completep then raise IncompleteSatisfiable;
      Proof.handle_result None

let search_main () =
  match (!mode) with
  | [] ->
      let starttime = Unix.gettimeofday () in
      Read.read_problem !problemfile; (*** parse the problem before calling slaves - Chad, Nov 2016 ***)
      let schedule = Schedule.get !schedule_files in
      let usertime =
        begin match !timeout with
          None -> None
        | Some timeout -> Some (starttime +. timeout)
        end in
      run_schedule usertime (schedule_duration schedule) schedule
  | modes -> run_modes modes
